import logging
import os
import subprocess
from collections import namedtuple
from pathlib import Path
from typing import Generator, Iterable, Iterator, List

import numpy as np
import pandas as pd
import pytest
import xgboost as xgb
from pyspark import SparkConf
from pyspark.ml import Pipeline, PipelineModel
from pyspark.ml.evaluation import BinaryClassificationEvaluator
from pyspark.ml.feature import VectorAssembler
from pyspark.ml.functions import vector_to_array
from pyspark.ml.linalg import Vectors
from pyspark.ml.tuning import CrossValidator, ParamGridBuilder
from pyspark.sql import SparkSession
from pyspark.sql import functions as spark_sql_func
from xgboost import XGBClassifier, XGBRegressor
from xgboost import testing as tm
from xgboost.callback import LearningRateScheduler
from xgboost.collective import Config
from xgboost.spark import (
    SparkXGBClassifier,
    SparkXGBClassifierModel,
    SparkXGBRanker,
    SparkXGBRegressor,
    SparkXGBRegressorModel,
)
from xgboost.spark.utils import _get_max_num_concurrent_tasks
from xgboost.testing.collective import get_avail_port

logging.getLogger("py4j").setLevel(logging.INFO)

pytestmark = [tm.timeout(60), pytest.mark.skipif(**tm.no_spark())]


def _to_bool(value: object) -> bool:
    if isinstance(value, bool):
        return value
    return str(value).lower() in {"1", "true", "on", "yes"}


def _probe_gpu_addresses() -> List[str]:
    if tm.no_spark()["condition"]:
        return []

    info = xgb.build_info()
    use_cuda = _to_bool(info.get("USE_CUDA", False))
    if not use_cuda:
        return []

    try:
        completed = subprocess.run(
            ["nvidia-smi", "--query-gpu=index", "--format=csv,noheader"],
            capture_output=True,
            text=True,
            check=False,
        )
    except OSError:
        return []

    if completed.returncode != 0:
        return []

    addresses = [line.strip() for line in completed.stdout.splitlines() if line.strip()]
    return addresses


_GPU_ADDRESSES = _probe_gpu_addresses()
_NUM_GPUS = len(_GPU_ADDRESSES)
_GPU_DISCOVERY_SCRIPT = os.path.join(os.path.dirname(__file__), "discover_gpu.sh")
_HAS_GPU_SPARK_MODE = bool(_GPU_ADDRESSES)
_GPU_SKIP_REASON = (
    "local_cluster_gpu requires CUDA-enabled XGBoost and visible GPUs via nvidia-smi."
)

SPARK_MODES = [
    pytest.param("local", id="local"),
    pytest.param("local_cluster", id="local_cluster"),
    pytest.param("local_cluster_connect", id="local_cluster_connect"),
    pytest.param(
        "local_cluster_gpu",
        id="local_cluster_gpu",
        marks=[
            tm.timeout(240),
            pytest.mark.skipif(not _HAS_GPU_SPARK_MODE, reason=_GPU_SKIP_REASON),
        ],
    ),
]


def _spark_test_mode(spark: SparkSession) -> str:
    return spark.conf.get("spark.xgboost.test.mode")


def _spark_test_device(spark: SparkSession) -> str:
    if _spark_test_mode(spark) == "local_cluster_gpu":
        return "cuda"
    return "cpu"


@pytest.fixture(scope="module")
def spark(request: pytest.FixtureRequest) -> Generator[SparkSession, None, None]:
    mode = getattr(request, "param", "local")
    if mode not in {
        "local",
        "local_cluster",
        "local_cluster_connect",
        "local_cluster_gpu",
    }:
        raise ValueError(f"Unknown Spark test mode: {mode}")
    os.environ["XGBOOST_PYSPARK_SHARED_SESSION"] = "1"
    config = {
        "spark.master": "local[4]",
        "spark.python.worker.reuse": "true",
        "spark.driver.host": "127.0.0.1",
        "spark.task.maxFailures": "1",
        "spark.sql.shuffle.partitions": "4",
        "spark.sql.execution.pyspark.udf.simplifiedTraceback.enabled": "false",
        "spark.sql.pyspark.jvmStacktrace.enabled": "true",
        "spark.ui.enabled": "false",
        "spark.xgboost.test.mode": mode,
    }
    if mode == "local_cluster":
        config.update(
            {
                "spark.master": "local-cluster[2, 1, 1024]",
                "spark.cores.max": "2",
                "spark.task.cpus": "1",
                "spark.executor.cores": "1",
            }
        )
    elif mode == "local_cluster_connect":
        del config["spark.master"]
        config.update(
            {
                "spark.remote": "local-cluster[2, 1, 1024]",
                "spark.cores.max": "2",
                "spark.task.cpus": "1",
                "spark.executor.cores": "1",
            }
        )
    elif mode == "local_cluster_gpu":
        # Use all available GPUs with a single worker node.
        config.update(
            {
                "spark.master": f"local-cluster[1, {_NUM_GPUS}, 1024]",
                "spark.python.worker.reuse": "false",
                "spark.cores.max": str(_NUM_GPUS),
                "spark.task.cpus": "1",
                "spark.executor.cores": str(_NUM_GPUS),
                "spark.default.parallelism": str(_NUM_GPUS),
                "spark.worker.resource.gpu.amount": str(_NUM_GPUS),
                "spark.task.resource.gpu.amount": "1",
                "spark.executor.resource.gpu.amount": str(_NUM_GPUS),
                "spark.worker.resource.gpu.discoveryScript": _GPU_DISCOVERY_SCRIPT,
            }
        )

    builder = SparkSession.builder.appName("XGBoost PySpark Python API Tests")
    for k, v in config.items():
        builder.config(k, v)
    logging.getLogger("pyspark").setLevel(logging.INFO)
    sess = builder.getOrCreate()
    if mode in {"local_cluster", "local_cluster_connect", "local_cluster_gpu"}:
        # Block until workers are connected.
        num_slots = int(sess.conf.get("spark.cores.max", "2"))
        sess.range(num_slots).repartition(num_slots).count()
    try:
        yield sess
    finally:
        sess.stop()
        os.environ.pop("XGBOOST_PYSPARK_SHARED_SESSION", None)


@pytest.fixture(scope="module")
def num_workers(spark: SparkSession) -> int:
    if _spark_test_mode(spark) == "local_cluster_connect":
        return int(spark.conf.get("spark.cores.max"))
    if _spark_test_mode(spark) == "local_cluster_gpu":
        return _NUM_GPUS
    return _get_max_num_concurrent_tasks(spark)


RegData = namedtuple(
    "RegData",
    (
        "X_train",
        "X_test",
        "y_train",
        "y_test",
        "weights",
        "base_margin",
        "is_val",
        "X",
        "y",
        "df",
    ),
)


class TestRegressor:
    @pytest.fixture(scope="class")
    def reg_data(self, spark: SparkSession) -> RegData:
        rng = np.random.default_rng(seed=42)
        X = rng.random((100, 10))
        # Make odd rows sparse with some random values to test both dense and sparse paths.
        X[1::2, :] = 0.0
        X[1::2, 1] = rng.random(len(X[1::2, 1]))
        X[1::2, 2] = rng.random(len(X[1::2, 2]))
        y = rng.random(100)
        w = rng.random(100)
        base_margin = rng.random(100)
        is_val = rng.random(100) < 0.2
        X_train, X_test = X[~is_val], X[is_val]
        y_train, y_test = y[~is_val], y[is_val]
        rows = []
        for i in range(len(y)):
            vec = (
                Vectors.dense(X[i, :])
                if i % 2 == 0
                else Vectors.sparse(X.shape[1], {1: float(X[i, 1]), 2: float(X[i, 2])})
            )
            rows.append(
                (
                    i,
                    vec,
                    float(y[i]),
                    float(w[i]),
                    float(base_margin[i]),
                    bool(is_val[i]),
                )
            )
        df = spark.createDataFrame(
            rows, ["row_id", "features", "label", "weight", "base_margin", "is_val"]
        )
        return RegData(
            X_train, X_test, y_train, y_test, w, base_margin, is_val, X, y, df
        )

    @pytest.mark.parametrize("spark", SPARK_MODES, indirect=True)
    def test_regressor(
        self, spark: SparkSession, reg_data: RegData, num_workers: int
    ) -> None:
        device = _spark_test_device(spark)

        reg_param = {
            "n_estimators": 10,
            "max_depth": 5,
            "objective": "reg:squarederror",
            "max_bin": 9,
            "eval_metric": "rmse",
            "early_stopping_rounds": 1,
            "device": device,
        }
        spark_regressor = SparkXGBRegressor(
            pred_contrib_col="pred_contribs",
            weight_col="weight",
            validation_indicator_col="is_val",
            num_workers=num_workers,
            **reg_param,
        ).fit(reg_data.df)
        pred_result = spark_regressor.transform(reg_data.df)
        preds = (
            pred_result.orderBy("row_id")
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )
        pred_contribs = np.array(
            pred_result.orderBy("row_id")
            .select("pred_contribs")
            .toPandas()["pred_contribs"]
            .tolist()
        )
        iter_range = (0, 1)
        spark_iter_regressor = SparkXGBRegressor(
            weight_col="weight",
            validation_indicator_col="is_val",
            iteration_range=iter_range,
            num_workers=num_workers,
            **reg_param,
        ).fit(reg_data.df)
        iter_preds = (
            spark_iter_regressor.transform(reg_data.df)
            .orderBy("row_id")
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )

        train_history = spark_regressor.training_summary.train_objective_history["rmse"]
        valid_history = spark_regressor.training_summary.validation_objective_history[
            "rmse"
        ]
        assert len(train_history) > 0
        assert len(valid_history) > 0
        assert len(train_history) == len(valid_history)
        assert np.isfinite(train_history).all()
        assert np.isfinite(valid_history).all()
        assert preds.shape == (len(reg_data.y),)
        assert iter_preds.shape == preds.shape

        assert np.allclose(pred_contribs.sum(axis=1), preds, rtol=1e-3)
        assert np.allclose(
            min(valid_history),
            spark_regressor._xgb_sklearn_model.best_score,
        )

    def test_training_continuation(self, reg_data: RegData) -> None:
        params = {
            "max_depth": 3,
            "objective": "reg:squarederror",
            "eval_metric": "rmse",
        }

        base = SparkXGBRegressor(n_estimators=2, **params).fit(reg_data.df)
        continued = SparkXGBRegressor(
            n_estimators=4, xgb_model=base.get_booster(), **params
        ).fit(reg_data.df)

        preds_base = (
            base.transform(reg_data.df)
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )
        preds_cont = (
            continued.transform(reg_data.df)
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )

        ref_base = XGBRegressor(n_estimators=2, **params).fit(reg_data.X, reg_data.y)
        ref_cont = XGBRegressor(n_estimators=4, **params).fit(
            reg_data.X, reg_data.y, xgb_model=ref_base.get_booster()
        )

        assert np.allclose(preds_cont, ref_cont.predict(reg_data.X), rtol=1e-3)
        assert not np.allclose(preds_base, preds_cont, rtol=1e-6)

    def test_regressor_with_base_margin(self, reg_data: RegData) -> None:
        params = {
            "n_estimators": 5,
            "max_depth": 3,
            "objective": "reg:squarederror",
        }
        spark_model = SparkXGBRegressor(base_margin_col="base_margin", **params).fit(
            reg_data.df
        )
        preds = (
            spark_model.transform(
                reg_data.df.select("row_id", "features", "base_margin")
            )
            .orderBy("row_id")
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )

        ref = XGBRegressor(**params).fit(
            reg_data.X, reg_data.y, base_margin=reg_data.base_margin
        )
        expected = ref.predict(reg_data.X, base_margin=reg_data.base_margin)

        assert np.allclose(preds, expected, rtol=1e-3)

    def test_regressor_save_load(self, reg_data: RegData, tmp_path: Path) -> None:
        train_df = reg_data.df.select("features", "label")
        model = SparkXGBRegressor(n_estimators=5, max_depth=3).fit(train_df)
        preds_before = (
            model.transform(train_df)
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )

        path = str(tmp_path / "spark-xgb-reg-model")
        model.save(path)
        loaded = SparkXGBRegressorModel.load(path)
        preds_after = (
            loaded.transform(train_df)
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )

        assert np.allclose(preds_before, preds_after, rtol=1e-6)

    def test_regressor_params(self, spark: SparkSession) -> None:
        py_reg = SparkXGBRegressor()
        assert hasattr(py_reg, "n_estimators")
        assert py_reg.n_estimators.parent == py_reg.uid
        assert not hasattr(py_reg, "gpu_id")
        assert hasattr(py_reg, "device")
        assert py_reg.getOrDefault(py_reg.n_estimators) == 100
        assert py_reg.getOrDefault(getattr(py_reg, "objective")) == "reg:squarederror"
        py_reg2 = SparkXGBRegressor(n_estimators=200)
        assert py_reg2.getOrDefault(getattr(py_reg2, "n_estimators")) == 200
        py_reg3 = py_reg2.copy({getattr(py_reg2, "max_depth"): 10})
        assert py_reg3.getOrDefault(getattr(py_reg3, "n_estimators")) == 200
        assert py_reg3.getOrDefault(getattr(py_reg3, "max_depth")) == 10
        with pytest.raises(ValueError, match="Number of workers"):
            SparkXGBRegressor(num_workers=-1)._validate_params(spark)
        with pytest.raises(ValueError, match="Number of workers"):
            SparkXGBRegressor(num_workers=0)._validate_params(spark)

    def test_valid_type(self, spark: SparkSession) -> None:
        df_train = spark.createDataFrame(
            [
                (Vectors.dense(1.0, 2.0, 3.0), 0, 0),
                (Vectors.sparse(3, {1: 1.0, 2: 5.5}), 1, 0),
                (Vectors.dense(4.0, 5.0, 6.0), 0, 1),
                (Vectors.sparse(3, {1: 6.0, 2: 7.5}), 1, 1),
            ],
            ["features", "label", "isVal"],
        )
        reg = SparkXGBRegressor(
            features_col="features",
            label_col="label",
            validation_indicator_col="isVal",
        )
        with pytest.raises(TypeError, match="The validation indicator must be boolean"):
            reg.fit(df_train)

    @pytest.mark.parametrize("spark", SPARK_MODES, indirect=True)
    def test_callbacks(
        self, spark: SparkSession, reg_data: RegData, tmp_path: Path
    ) -> None:
        train_df = reg_data.df.select("row_id", "features", "label")
        device = _spark_test_device(spark)

        def custom_lr(boosting_round: int) -> float:
            return 1.0 / (boosting_round + 1)

        reg_params = {
            "n_estimators": 10,
            "max_depth": 3,
            "objective": "reg:squarederror",
            "eval_metric": "rmse",
            "device": device,
        }

        path = str(tmp_path / "spark-xgb-reg-cb")
        regressor = SparkXGBRegressor(
            callbacks=[LearningRateScheduler(custom_lr)], **reg_params
        )
        regressor.save(path)
        regressor = SparkXGBRegressor.load(path)
        loaded_callbacks = regressor.getOrDefault(regressor.callbacks)
        assert loaded_callbacks is not None
        assert len(loaded_callbacks) == 1

        model = regressor.fit(train_df)
        preds = (
            model.transform(train_df)
            .orderBy("row_id")
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )

        ref = XGBRegressor(
            callbacks=[LearningRateScheduler(custom_lr)], **reg_params
        ).fit(reg_data.X, reg_data.y)
        assert np.allclose(preds, ref.predict(reg_data.X), rtol=1e-3)

    @pytest.mark.parametrize("tree_method", ["hist", "approx"])
    def test_empty_train_data(self, spark: SparkSession, tree_method: str) -> None:
        df_train = spark.createDataFrame(
            [
                (Vectors.dense(10.1, 11.2, 11.3), 0, True),
                (Vectors.dense(1, 1.2, 1.3), 1, True),
                (Vectors.dense(14.0, 15.0, 16.0), 0, True),
                (Vectors.dense(1.1, 1.2, 1.3), 1, False),
            ],
            ["features", "label", "val_col"],
        )
        classifier = SparkXGBRegressor(
            num_workers=2,
            min_child_weight=0.0,
            reg_alpha=0,
            reg_lambda=0,
            tree_method=tree_method,
            validation_indicator_col="val_col",
        )
        model = classifier.fit(df_train)
        pred_result = model.transform(df_train).collect()
        for row in pred_result:
            assert row.prediction == 1.0


ClfData = namedtuple(
    "ClfData",
    (
        "X_train",
        "X_test",
        "y_train",
        "y_test",
        "weights",
        "base_margin",
        "is_val",
        "X",
        "y",
        "df",
    ),
)


class TestClassifier:
    @pytest.fixture(scope="class")
    def clf_data(self, spark: SparkSession) -> ClfData:
        rng = np.random.default_rng(seed=123)
        X = rng.random((200, 10))
        X[1::2, :] = 0.0
        X[1::2, 1] = rng.random(len(X[1::2, 1]))
        X[1::2, 2] = rng.random(len(X[1::2, 2]))
        y = rng.integers(0, 2, size=200)
        w = rng.random(200)
        base_margin = rng.random(200)
        is_val = rng.random(200) < 0.2
        X_train, X_test = X[~is_val], X[is_val]
        y_train, y_test = y[~is_val], y[is_val]
        rows = []
        for i in range(len(y)):
            vec = (
                Vectors.dense(X[i, :])
                if i % 2 == 0
                else Vectors.sparse(X.shape[1], {1: float(X[i, 1]), 2: float(X[i, 2])})
            )
            rows.append(
                (
                    i,
                    vec,
                    int(y[i]),
                    float(w[i]),
                    float(base_margin[i]),
                    bool(is_val[i]),
                )
            )
        df = spark.createDataFrame(
            rows, ["row_id", "features", "label", "weight", "base_margin", "is_val"]
        )
        return ClfData(
            X_train, X_test, y_train, y_test, w, base_margin, is_val, X, y, df
        )

    @pytest.mark.parametrize("spark", SPARK_MODES, indirect=True)
    def test_classifier(
        self, spark: SparkSession, clf_data: ClfData, num_workers: int
    ) -> None:
        train_df = clf_data.df
        X = clf_data.X
        y = clf_data.y
        weights = clf_data.weights
        train_rows = np.where(~clf_data.is_val)[0]
        validation_rows = np.where(clf_data.is_val)[0]
        device = _spark_test_device(spark)

        cls_params = {
            "n_estimators": 10,
            "max_depth": 5,
            "eval_metric": "logloss",
            "device": device,
        }
        ref = XGBClassifier(**cls_params).fit(
            X[train_rows],
            y[train_rows],
            sample_weight=weights[train_rows],
            eval_set=[
                (X[train_rows], y[train_rows]),
                (X[validation_rows], y[validation_rows]),
            ],
            sample_weight_eval_set=[weights[train_rows], weights[validation_rows]],
        )

        spark_cls = SparkXGBClassifier(
            weight_col="weight",
            validation_indicator_col="is_val",
            num_workers=num_workers,
            **cls_params,
        ).fit(train_df)

        pred_result = spark_cls.transform(train_df)
        preds = (
            pred_result.orderBy("row_id")
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )
        proba = np.array(
            pred_result.orderBy("row_id")
            .select("probability")
            .toPandas()["probability"]
            .tolist()
        )

        assert preds.shape == ref.predict(X).shape
        assert proba.shape == ref.predict_proba(X).shape
        assert np.allclose(
            ref.evals_result()["validation_0"]["logloss"],
            spark_cls.training_summary.train_objective_history["logloss"],
            atol=2e-2,
        )
        assert np.allclose(
            ref.evals_result()["validation_1"]["logloss"],
            spark_cls.training_summary.validation_objective_history["logloss"],
            atol=2e-2,
        )

    def test_classifier_model_save_load(
        self, clf_data: ClfData, tmp_path: Path
    ) -> None:
        train_df = clf_data.df.select("features", "label")
        test_df = clf_data.df.select("features")
        path = str(tmp_path / "spark-xgb-clf-model")
        clf = SparkXGBClassifier(n_estimators=5, max_depth=3)
        model = clf.fit(train_df)
        model.save(path)
        loaded_model = SparkXGBClassifierModel.load(path)
        assert model.uid == loaded_model.uid
        pred_before = (
            model.transform(test_df)
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )
        pred_after = (
            loaded_model.transform(test_df)
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )
        assert np.allclose(pred_before, pred_after, rtol=1e-6)

        with pytest.raises(AssertionError, match="Expected class name"):
            SparkXGBRegressorModel.load(path)

    def test_classifier_model_pipeline_save_load(
        self, clf_data: ClfData, tmp_path: Path
    ) -> None:
        train_df = clf_data.df.select("features", "label")
        test_df = clf_data.df.select("features")
        path = str(tmp_path / "spark-xgb-clf-pipeline")
        classifier = SparkXGBClassifier()
        pipeline = Pipeline(stages=[classifier])
        pipeline = pipeline.copy(
            extra={
                getattr(classifier, k): v
                for k, v in {"max_depth": 5, "n_estimators": 10}.items()
            }
        )
        model = pipeline.fit(train_df)
        model.save(path)

        loaded_model = PipelineModel.load(path)
        pred_before = (
            model.transform(test_df)
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )
        pred_after = (
            loaded_model.transform(test_df)
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )
        assert np.allclose(pred_before, pred_after, rtol=1e-6)

    def test_classifier_params(self, spark: SparkSession) -> None:
        py_clf = SparkXGBClassifier()
        assert hasattr(py_clf, "n_estimators")
        assert py_clf.n_estimators.parent == py_clf.uid
        assert not hasattr(py_clf, "gpu_id")
        assert hasattr(py_clf, "device")

        assert py_clf.getOrDefault(py_clf.n_estimators) == 100
        assert py_clf.getOrDefault(getattr(py_clf, "objective")) is None
        py_clf2 = SparkXGBClassifier(n_estimators=200)
        assert py_clf2.getOrDefault(getattr(py_clf2, "n_estimators")) == 200
        py_clf3 = py_clf2.copy({getattr(py_clf2, "max_depth"): 10})
        assert py_clf3.getOrDefault(getattr(py_clf3, "n_estimators")) == 200
        assert py_clf3.getOrDefault(getattr(py_clf3, "max_depth")) == 10
        with pytest.raises(ValueError, match="custom 'objective'"):
            SparkXGBClassifier(objective="binary:logistic")._validate_params(spark)
        assert hasattr(py_clf, "arbitrary_params_dict")
        assert py_clf.getOrDefault(py_clf.arbitrary_params_dict) == {}

        # Testing overwritten params via setParams
        py_clf_overwrite = SparkXGBClassifier()
        py_clf_overwrite.setParams(x=1, y=2)
        py_clf_overwrite.setParams(y=3, z=4)
        xgb_params = py_clf_overwrite._gen_xgb_params_dict()
        assert xgb_params["x"] == 1
        assert xgb_params["y"] == 3
        assert xgb_params["z"] == 4
        with pytest.raises(ValueError, match="evals_result"):
            SparkXGBClassifier(evals_result={})

    @pytest.mark.parametrize("tree_method", ["hist", "approx"])
    def test_empty_validation_data(self, spark: SparkSession, tree_method: str) -> None:
        df_train = spark.createDataFrame(
            [
                (Vectors.dense(10.1, 11.2, 11.3), 0, False),
                (Vectors.dense(1, 1.2, 1.3), 1, False),
                (Vectors.dense(14.0, 15.0, 16.0), 0, False),
                (Vectors.dense(1.1, 1.2, 1.3), 1, True),
            ],
            ["features", "label", "val_col"],
        )
        classifier = SparkXGBClassifier(
            num_workers=2,
            tree_method=tree_method,
            min_child_weight=0.0,
            reg_alpha=0,
            reg_lambda=0,
            validation_indicator_col="val_col",
            n_estimators=10,
        )
        model = classifier.fit(df_train)
        pred_result = model.transform(df_train).collect()
        for row in pred_result:
            assert row.prediction == row.label

    @pytest.mark.parametrize("tree_method", ["hist", "approx"])
    def test_empty_partition(self, spark: SparkSession, tree_method: str) -> None:
        # raw_df.repartition(4) will result int severe data skew, actually,
        # there is no any data in reducer partition 1, reducer partition 2
        # see https://github.com/dmlc/xgboost/issues/8221
        raw_df = spark.range(0, 40, 1, 50).withColumn(
            "label",
            spark_sql_func.when(spark_sql_func.rand(1) > 0.5, 1).otherwise(0),
        )
        vector_assembler = (
            VectorAssembler().setInputCols(["id"]).setOutputCol("features")
        )
        data_trans = vector_assembler.setHandleInvalid("keep").transform(raw_df)
        classifier = SparkXGBClassifier(
            tree_method=tree_method,
            n_estimators=10,
        )
        model = classifier.fit(data_trans)
        pred_result = model.transform(data_trans).collect()
        for row in pred_result:
            assert row.prediction in [0.0, 1.0]

    def test_classifier_with_cross_validator(self, clf_data: ClfData) -> None:
        xgb_classifier = SparkXGBClassifier(n_estimators=1)
        param_maps = (
            ParamGridBuilder().addGrid(xgb_classifier.max_depth, [1, 2]).build()
        )
        cv_bin = CrossValidator(
            estimator=xgb_classifier,
            estimatorParamMaps=param_maps,
            evaluator=BinaryClassificationEvaluator(),
            seed=1,
            parallelism=4,
            numFolds=2,
        )
        cv_model = cv_bin.fit(clf_data.df.select("features", "label"))
        cv_model.transform(clf_data.df.select("features"))

    def test_convert_to_sklearn_model_clf(self, clf_data: ClfData) -> None:
        classifier = SparkXGBClassifier(
            n_estimators=10,
            missing=2.0,
            max_depth=3,
            sketch_eps=0.5,
        )
        clf_model = classifier.fit(clf_data.df.select("features", "label"))

        # Check that regardless of what booster, _convert_to_model converts to the
        # correct class type
        sklearn_classifier = classifier._convert_to_sklearn_model(
            clf_model.get_booster().save_raw("json"),
            clf_model.get_booster().save_config(),
        )
        assert isinstance(sklearn_classifier, XGBClassifier)
        assert sklearn_classifier.n_estimators == 10
        assert sklearn_classifier.missing == 2.0
        assert sklearn_classifier.max_depth == 3
        assert sklearn_classifier.get_params()["sketch_eps"] == 0.5

    def test_classifier_array_col_as_feature(self, clf_data: ClfData) -> None:
        vector_train = clf_data.df.select("row_id", "features", "label")
        vector_test = clf_data.df.select("row_id", "features")
        train_dataset = vector_train.withColumn(
            "features", vector_to_array(spark_sql_func.col("features"))
        )
        test_dataset = vector_test.withColumn(
            "features", vector_to_array(spark_sql_func.col("features"))
        )
        params = {"n_estimators": 10, "max_depth": 3}
        vector_model = SparkXGBClassifier(**params).fit(vector_train)
        array_model = SparkXGBClassifier(**params).fit(train_dataset)

        vector_pred = (
            vector_model.transform(vector_test)
            .orderBy("row_id")
            .select("prediction", "probability")
            .toPandas()
        )
        array_pred = (
            array_model.transform(test_dataset)
            .orderBy("row_id")
            .select("prediction", "probability")
            .toPandas()
        )
        vector_proba = np.array(vector_pred["probability"].tolist())
        array_proba = np.array(array_pred["probability"].tolist())
        array_label = array_pred["prediction"].to_numpy()

        assert np.allclose(
            array_pred["prediction"].to_numpy(), vector_pred["prediction"].to_numpy()
        )
        assert np.allclose(array_proba, vector_proba, rtol=1e-3)
        assert np.allclose(
            array_proba.sum(axis=1), np.ones(array_proba.shape[0]), atol=1e-6
        )
        assert np.all((array_proba >= 0.0) & (array_proba <= 1.0))
        assert np.allclose(array_label, np.argmax(array_proba, axis=1))

    def test_classifier_with_feature_names_types(self, clf_data: ClfData) -> None:
        n_features = clf_data.X.shape[1]
        classifier = SparkXGBClassifier(
            feature_names=[f"f{i}" for i in range(n_features)],
            feature_types=["float"] * n_features,
            feature_weights=[float(i + 1) for i in range(n_features)],
            n_estimators=10,
        )
        model = classifier.fit(clf_data.df.select("features", "label"))
        model.transform(clf_data.df.select("features")).collect()

    def test_early_stop_param_validation(self, clf_data: ClfData) -> None:
        classifier = SparkXGBClassifier(early_stopping_rounds=1)
        with pytest.raises(ValueError, match="early_stopping_rounds"):
            classifier.fit(clf_data.df.select("features", "label"))

    def test_classifier_with_list_eval_metric(self, clf_data: ClfData) -> None:
        classifier = SparkXGBClassifier(eval_metric=["auc", "rmse"], n_estimators=10)
        model = classifier.fit(clf_data.df.select("features", "label"))
        model.transform(clf_data.df.select("features")).collect()

    def test_classifier_with_sparse_optim(self, spark: SparkSession) -> None:
        sparse_train = spark.createDataFrame(
            [
                (Vectors.dense(1.0, 0.0, 3.0, 0.0, 0.0), 0),
                (Vectors.sparse(5, {1: 1.0, 3: 5.5}), 1),
                (Vectors.sparse(5, {4: -3.0}), 0),
            ]
            * 5,
            ["features", "label"],
        )
        cls = SparkXGBClassifier(missing=0.0, n_estimators=10)
        model = cls.fit(sparse_train)
        assert model._xgb_sklearn_model.missing == 0.0
        pred_result = model.transform(sparse_train).collect()

        # enable sparse optimization
        cls2 = SparkXGBClassifier(
            missing=0.0,
            enable_sparse_data_optim=True,
            n_estimators=10,
        )
        model2 = cls2.fit(sparse_train)
        assert model2.getOrDefault(model2.enable_sparse_data_optim)
        assert model2._xgb_sklearn_model.missing == 0.0
        pred_result2 = model2.transform(sparse_train).collect()

        for row1, row2 in zip(pred_result, pred_result2):
            assert np.allclose(row1.probability, row2.probability, rtol=1e-3)

    def test_param_alias(self) -> None:
        py_cls = SparkXGBClassifier(features_col="f1", label_col="l1")
        assert py_cls.getOrDefault(py_cls.featuresCol) == "f1"
        assert py_cls.getOrDefault(py_cls.labelCol) == "l1"
        with pytest.raises(
            ValueError, match="Please use param name features_col instead"
        ):
            SparkXGBClassifier(featuresCol="f1")

    def test_param_value_converter(self) -> None:
        py_cls = SparkXGBClassifier(missing=np.float64(1.0), sketch_eps=np.float64(0.3))
        # don't check by isinstance(v, float) because for numpy scalar it will also return True
        assert py_cls.getOrDefault(py_cls.missing).__class__.__name__ == "float"
        assert (
            py_cls.getOrDefault(py_cls.arbitrary_params_dict)[
                "sketch_eps"
            ].__class__.__name__
            == "float64"
        )

    def test_device_and_gpu_params(
        self, clf_data: ClfData, spark: SparkSession
    ) -> None:
        clf = SparkXGBClassifier(device="cuda", tree_method="exact")
        with pytest.raises(ValueError, match="not supported for distributed"):
            clf.fit(clf_data.df.select("features", "label"))

        clf = SparkXGBClassifier(device="cuda", tree_method="approx")
        clf._validate_params(spark)
        clf = SparkXGBClassifier(device="cuda")
        clf._validate_params(spark)

        clf = SparkXGBClassifier()
        assert not clf._run_on_gpu(spark)

        clf = SparkXGBClassifier(device="cuda", tree_method="hist")
        assert clf._run_on_gpu(spark)

        clf = SparkXGBClassifier(device="cuda")
        assert clf._run_on_gpu(spark)

        clf = SparkXGBClassifier(tree_method="hist")
        assert not clf._run_on_gpu(spark)

        clf = SparkXGBClassifier(device="cuda", tree_method="approx")
        assert clf._run_on_gpu(spark)

    def test_gpu_transform(
        self, clf_data: ClfData, tmp_path: Path, spark: SparkSession
    ) -> None:
        """local mode"""
        classifier = SparkXGBClassifier(device="cpu", n_estimators=10)
        model: SparkXGBClassifierModel = classifier.fit(
            clf_data.df.select("features", "label")
        )

        path = "file:" + str(tmp_path)
        model.write().overwrite().save(path)

        # The model trained with CPU, transform defaults to cpu
        assert not model._run_on_gpu(spark)

        # without error
        model.transform(clf_data.df.select("features")).collect()

        model.set_device("cuda")
        assert model._run_on_gpu(spark)

        model_loaded = SparkXGBClassifierModel.load(path)

        # The model trained with CPU, transform defaults to cpu
        assert not model_loaded._run_on_gpu(spark)
        # without error
        model_loaded.transform(clf_data.df.select("features")).collect()

        model_loaded.set_device("cuda")
        assert model_loaded._run_on_gpu(spark)

    def test_validate_gpu_params(self) -> None:
        from unittest.mock import Mock

        def _mock_ss(version: str, conf: SparkConf) -> Mock:
            return Mock(version=version, conf=conf)

        # Standalone
        standalone_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.executor.cores", "12")
            .set("spark.task.cpus", "1")
            .set("spark.executor.resource.gpu.amount", "1")
            .set("spark.task.resource.gpu.amount", "0.08")
        )
        classifier_on_cpu = SparkXGBClassifier(device="cpu")
        classifier_on_gpu = SparkXGBClassifier(device="cuda")

        # No exception for classifier on CPU
        classifier_on_cpu._validate_gpu_params(_mock_ss("4.0.0", standalone_conf))

        # No issues for gpu configuration
        classifier_on_gpu._validate_gpu_params(_mock_ss("4.0.0", standalone_conf))

        # no spark.executor.resource.gpu.amount
        standalone_bad_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.executor.cores", "12")
            .set("spark.task.cpus", "1")
            .set("spark.task.resource.gpu.amount", "0.08")
        )
        msg_match = (
            "The `spark.executor.resource.gpu.amount` is required for training on GPU"
        )
        with pytest.raises(ValueError, match=msg_match):
            classifier_on_gpu._validate_gpu_params(
                _mock_ss("4.0.0", standalone_bad_conf)
            )

        # Without spark.task.resource.gpu.amount shouldn't throw error
        # since stage-level scheduling is available and will handle it.
        standalone_bad_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.executor.cores", "12")
            .set("spark.task.cpus", "1")
            .set("spark.executor.resource.gpu.amount", "1")
        )
        classifier_on_gpu._validate_gpu_params(_mock_ss("4.0.0", standalone_bad_conf))

        # Yarn and K8s mode
        for mode in ["yarn", "k8s://"]:
            conf = (
                SparkConf()
                .setMaster(mode)
                .set("spark.executor.cores", "12")
                .set("spark.task.cpus", "1")
                .set("spark.executor.resource.gpu.amount", "1")
                .set("spark.task.resource.gpu.amount", "0.08")
            )
            classifier_on_gpu._validate_gpu_params(_mock_ss("4.0.0", conf))

        for mode in ["yarn", "k8s://"]:
            bad_conf = (
                SparkConf()
                .setMaster(mode)
                .set("spark.executor.cores", "12")
                .set("spark.task.cpus", "1")
                .set("spark.executor.resource.gpu.amount", "1")
            )
            classifier_on_gpu._validate_gpu_params(_mock_ss("4.0.0", bad_conf))

    def test_skip_stage_level_scheduling(self) -> None:
        from unittest.mock import Mock

        def _mock_ss(version: str, conf: SparkConf) -> Mock:
            return Mock(version=version, conf=conf)

        standalone_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.executor.cores", "12")
            .set("spark.task.cpus", "1")
            .set("spark.executor.resource.gpu.amount", "1")
            .set("spark.task.resource.gpu.amount", "0.08")
        )

        classifier_on_cpu = SparkXGBClassifier(device="cpu")
        classifier_on_gpu = SparkXGBClassifier(device="cuda")

        # the correct configurations should not skip stage-level scheduling
        assert not classifier_on_gpu._skip_stage_level_scheduling(
            _mock_ss("4.0.0", standalone_conf)
        )

        # not run on GPU
        assert classifier_on_cpu._skip_stage_level_scheduling(
            _mock_ss("4.0.0", standalone_conf)
        )

        # spark.executor.cores is not set
        bad_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.task.cpus", "1")
            .set("spark.executor.resource.gpu.amount", "1")
            .set("spark.task.resource.gpu.amount", "0.08")
        )
        assert classifier_on_gpu._skip_stage_level_scheduling(
            _mock_ss("4.0.0", bad_conf)
        )

        # spark.executor.cores=1
        bad_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.executor.cores", "1")
            .set("spark.task.cpus", "1")
            .set("spark.executor.resource.gpu.amount", "1")
            .set("spark.task.resource.gpu.amount", "0.08")
        )
        assert classifier_on_gpu._skip_stage_level_scheduling(
            _mock_ss("4.0.0", bad_conf)
        )

        # spark.executor.resource.gpu.amount is not set
        bad_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.executor.cores", "12")
            .set("spark.task.cpus", "1")
            .set("spark.task.resource.gpu.amount", "0.08")
        )
        assert classifier_on_gpu._skip_stage_level_scheduling(
            _mock_ss("4.0.0", bad_conf)
        )

        # spark.executor.resource.gpu.amount>1
        bad_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.executor.cores", "12")
            .set("spark.task.cpus", "1")
            .set("spark.executor.resource.gpu.amount", "2")
            .set("spark.task.resource.gpu.amount", "0.08")
        )
        assert classifier_on_gpu._skip_stage_level_scheduling(
            _mock_ss("4.0.0", bad_conf)
        )

        # spark.task.resource.gpu.amount is not set
        bad_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.executor.cores", "12")
            .set("spark.task.cpus", "1")
            .set("spark.executor.resource.gpu.amount", "1")
        )
        assert not classifier_on_gpu._skip_stage_level_scheduling(
            _mock_ss("4.0.0", bad_conf)
        )

        # spark.task.resource.gpu.amount=1
        bad_conf = (
            SparkConf()
            .setMaster("spark://foo")
            .set("spark.executor.cores", "12")
            .set("spark.task.cpus", "1")
            .set("spark.executor.resource.gpu.amount", "1")
            .set("spark.task.resource.gpu.amount", "1")
        )
        assert classifier_on_gpu._skip_stage_level_scheduling(
            _mock_ss("4.0.0", bad_conf)
        )

        # For Yarn and K8S
        for mode in ["yarn", "k8s://"]:
            for gpu_amount in ["0.08", "0.2", "1.0"]:
                conf = (
                    SparkConf()
                    .setMaster(mode)
                    .set("spark.executor.cores", "12")
                    .set("spark.task.cpus", "1")
                    .set("spark.executor.resource.gpu.amount", "1")
                    .set("spark.task.resource.gpu.amount", gpu_amount)
                )

                if gpu_amount == "1.0":
                    assert classifier_on_gpu._skip_stage_level_scheduling(
                        _mock_ss("4.0.0", conf)
                    )
                else:
                    assert not classifier_on_gpu._skip_stage_level_scheduling(
                        _mock_ss("4.0.0", conf)
                    )

    def test_collective_conf(self, spark: SparkSession, tmp_path: Path) -> None:
        classifier = SparkXGBClassifier(
            launch_tracker_on_driver=True,
            coll_cfg=Config(tracker_host_ip="192.168.1.32", tracker_port=59981),
        )
        with pytest.raises(Exception, match="Failed to bind socket"):
            classifier._get_tracker_args(spark)

        classifier = SparkXGBClassifier(
            launch_tracker_on_driver=False,
            coll_cfg=Config(tracker_host_ip="127.0.0.1", tracker_port=58892),
        )
        with pytest.raises(
            ValueError, match="You must enable launch_tracker_on_driver"
        ):
            classifier._get_tracker_args(spark)

        avail_tracker_port = get_avail_port()
        classifier = SparkXGBClassifier(
            launch_tracker_on_driver=True,
            coll_cfg=Config(
                tracker_host_ip="127.0.0.1", tracker_port=avail_tracker_port
            ),
            num_workers=2,
        )
        launch_tracker_on_driver, rabit_envs = classifier._get_tracker_args(spark)
        assert launch_tracker_on_driver is True
        assert rabit_envs["n_workers"] == 2
        assert rabit_envs["dmlc_tracker_uri"] == "127.0.0.1"
        assert rabit_envs["dmlc_tracker_port"] == avail_tracker_port

        path = "file:" + str(tmp_path)
        port = get_avail_port()
        classifier = SparkXGBClassifier(
            launch_tracker_on_driver=True,
            coll_cfg=Config(tracker_host_ip="127.0.0.1", tracker_port=port),
            num_workers=1,
            n_estimators=1,
        )

        def check_conf(conf: Config) -> None:
            assert conf.tracker_host_ip == "127.0.0.1"
            assert conf.tracker_port == port

        check_conf(classifier.getOrDefault(classifier.coll_cfg))
        # PySpark Connect ML does not support overwrite - this is a bug in Spark:
        # https://issues.apache.org/jira/browse/SPARK-55452
        if _spark_test_mode(spark) == "local_cluster_connect":
            classifier.write().save(path)
        else:
            classifier.write().overwrite().save(path)

        loaded_classifier = SparkXGBClassifier.load(path)
        check_conf(loaded_classifier.getOrDefault(loaded_classifier.coll_cfg))

        sparse_train = spark.createDataFrame(
            [
                (Vectors.dense(1.0, 0.0, 3.0, 0.0, 0.0), 0),
                (Vectors.sparse(5, {1: 1.0, 3: 5.5}), 1),
                (Vectors.sparse(5, {4: -3.0}), 0),
            ]
            * 5,
            ["features", "label"],
        )
        model = classifier.fit(sparse_train)
        check_conf(model.getOrDefault(model.coll_cfg))
        # PySpark ML Connect does not support overwrite - this is a bug in Spark:
        # https://issues.apache.org/jira/browse/SPARK-55452
        if _spark_test_mode(spark) == "local_cluster_connect":
            model.write().save(path)
        else:
            model.write().overwrite().save(path)
        loaded_model = SparkXGBClassifierModel.load(path)
        check_conf(loaded_model.getOrDefault(loaded_model.coll_cfg))

    def test_launch_tracker_on_driver_initialization(self, spark: SparkSession) -> None:
        from unittest.mock import patch

        from xgboost.spark.utils import _is_connect

        with patch("xgboost.spark.core._get_rabit_args", return_value={}):
            # 1. Not set explicitly, no default
            clf1 = SparkXGBClassifier()
            launch_tracker1, _ = clf1._get_tracker_args(spark)
            assert launch_tracker1 is not _is_connect(spark)

            # 2. Set explicitly to True
            clf2 = SparkXGBClassifier(launch_tracker_on_driver=True)
            launch_tracker2, _ = clf2._get_tracker_args(spark)
            assert launch_tracker2 is True

            # 3. Set explicitly to False
            clf3 = SparkXGBClassifier(launch_tracker_on_driver=False)
            launch_tracker3, _ = clf3._get_tracker_args(spark)
            assert launch_tracker3 is False

            # 4. Set via default
            clf4 = SparkXGBClassifier()
            clf4._setDefault(launch_tracker_on_driver=False)
            launch_tracker4, _ = clf4._get_tracker_args(spark)
            assert launch_tracker4 is False

            clf5 = SparkXGBClassifier()
            clf5._setDefault(launch_tracker_on_driver=True)
            launch_tracker5, _ = clf5._get_tracker_args(spark)
            assert launch_tracker5 is True


LTRData = namedtuple(
    "LTRData",
    (
        "ranker_df",
        "X_train",
        "y_train",
        "qid_train",
        "X_test",
        "y_test",
        "qid_test",
    ),
)


class TestPySparkLocalLETOR:
    @pytest.fixture(scope="class")
    def ltr_data(self, spark: SparkSession) -> LTRData:
        spark.conf.set("spark.sql.execution.arrow.maxRecordsPerBatch", "8")
        ranker_df = spark.createDataFrame(
            [
                (Vectors.dense(1.0, 2.0, 3.0), 0, 0, None, False),
                (Vectors.dense(4.0, 5.0, 6.0), 1, 0, None, False),
                (Vectors.dense(9.0, 4.0, 8.0), 2, 0, None, False),
                (Vectors.sparse(3, {1: 1.0, 2: 5.5}), 0, 1, None, False),
                (Vectors.sparse(3, {1: 6.0, 2: 7.5}), 1, 1, None, False),
                (Vectors.sparse(3, {1: 8.0, 2: 9.5}), 2, 1, None, False),
                (Vectors.dense(1.5, 2.0, 3.0), 1, 0, 0, True),
                (Vectors.dense(4.5, 5.0, 6.0), 0, 0, 1, True),
                (Vectors.dense(9.0, 4.5, 8.0), 2, 0, 2, True),
                (Vectors.sparse(3, {1: 1.0, 2: 6.0}), 1, 1, 3, True),
                (Vectors.sparse(3, {1: 6.0, 2: 7.0}), 1, 1, 4, True),
                (Vectors.sparse(3, {1: 8.0, 2: 10.5}), 2, 1, 5, True),
            ],
            ["features", "label", "qid", "row_id", "isVal"],
        )
        X_train = np.array(
            [
                [1.0, 2.0, 3.0],
                [4.0, 5.0, 6.0],
                [9.0, 4.0, 8.0],
                [np.nan, 1.0, 5.5],
                [np.nan, 6.0, 7.5],
                [np.nan, 8.0, 9.5],
            ]
        )
        qid_train = np.array([0, 0, 0, 1, 1, 1])
        y_train = np.array([0, 1, 2, 0, 1, 2])

        X_test = np.array(
            [
                [1.5, 2.0, 3.0],
                [4.5, 5.0, 6.0],
                [9.0, 4.5, 8.0],
                [np.nan, 1.0, 6.0],
                [np.nan, 6.0, 7.0],
                [np.nan, 8.0, 10.5],
            ]
        )
        qid_test = np.array([0, 0, 0, 1, 1, 1])
        y_test = np.array([1, 0, 2, 1, 1, 2])

        return LTRData(
            ranker_df,
            X_train,
            y_train,
            qid_train,
            X_test,
            y_test,
            qid_test,
        )

    def test_ranker(self, ltr_data: LTRData) -> None:
        ref = xgb.XGBRanker(
            tree_method="approx",
            objective="rank:pairwise",
            n_estimators=10,
        )
        ref.fit(
            ltr_data.X_train,
            ltr_data.y_train,
            qid=ltr_data.qid_train,
            eval_set=[(ltr_data.X_test, ltr_data.y_test)],
            eval_qid=[ltr_data.qid_test],
        )
        expected = ref.predict(ltr_data.X_test)

        ranker = SparkXGBRanker(
            qid_col="qid",
            tree_method="approx",
            objective="rank:pairwise",
            validation_indicator_col="isVal",
            n_estimators=10,
        )
        assert ranker.getOrDefault(ranker.objective) == "rank:pairwise"
        model = ranker.fit(ltr_data.ranker_df)
        test_df = ltr_data.ranker_df.where(spark_sql_func.col("isVal"))
        pred_result = (
            model.transform(test_df)
            .orderBy("row_id")
            .select("prediction")
            .toPandas()["prediction"]
            .to_numpy()
        )
        assert np.allclose(pred_result, expected, rtol=1e-3)

    def test_ranker_same_qid_in_same_partition(self, spark: SparkSession) -> None:
        ranker_df_train = spark.createDataFrame(
            [
                (Vectors.sparse(3, {1: 1.0, 2: 5.5}), 0, 9),
                (Vectors.sparse(3, {1: 6.0, 2: 7.5}), 1, 9),
                (Vectors.sparse(3, {1: 8.0, 2: 9.5}), 2, 9),
                (Vectors.dense(1.0, 2.0, 3.0), 0, 8),
                (Vectors.dense(4.0, 5.0, 6.0), 1, 8),
                (Vectors.dense(9.0, 4.0, 8.0), 2, 8),
                (Vectors.sparse(3, {1: 1.0, 2: 5.5}), 0, 7),
                (Vectors.sparse(3, {1: 6.0, 2: 7.5}), 1, 7),
                (Vectors.sparse(3, {1: 8.0, 2: 9.5}), 2, 7),
                (Vectors.dense(1.0, 2.0, 3.0), 0, 6),
                (Vectors.dense(4.0, 5.0, 6.0), 1, 6),
                (Vectors.dense(9.0, 4.0, 8.0), 2, 6),
            ]
            * 4,
            ["features", "label", "qid"],
        )
        ranker = SparkXGBRanker(qid_col="qid", num_workers=4, force_repartition=True)
        df, _ = ranker._prepare_input(ranker_df_train)

        def f(iterator: Iterator[pd.DataFrame]) -> Iterator[pd.DataFrame]:
            unique_qids = set()
            for pdf in iterator:
                unique_qids.update(pdf["qid"].unique())
            yield pd.DataFrame({"qids": [list(unique_qids)]})

        rows = df.select("qid").mapInPandas(f, schema="qids array<int>").collect()
        assert len(rows) == 4
        for row in rows:
            assert len(row.qids) == 1
            assert row.qids[0] in [6, 7, 8, 9]

    def test_ranker_xgb_summary(self, ltr_data: LTRData) -> None:
        spark_xgb_model = SparkXGBRanker(
            tree_method="approx",
            qid_col="qid",
            objective="rank:pairwise",
            validation_indicator_col="isVal",
            n_estimators=10,
        ).fit(ltr_data.ranker_df)

        ref = xgb.XGBRanker(
            tree_method="approx",
            objective="rank:pairwise",
            n_estimators=10,
        )
        ref.fit(
            ltr_data.X_train,
            ltr_data.y_train,
            qid=ltr_data.qid_train,
            eval_set=[
                (ltr_data.X_train, ltr_data.y_train),
                (ltr_data.X_test, ltr_data.y_test),
            ],
            eval_qid=[ltr_data.qid_train, ltr_data.qid_test],
        )

        np.testing.assert_allclose(
            ref.evals_result()["validation_0"]["ndcg@32"],
            spark_xgb_model.training_summary.train_objective_history["ndcg@32"],
            atol=1e-3,
        )

        np.testing.assert_allclose(
            ref.evals_result()["validation_1"]["ndcg@32"],
            spark_xgb_model.training_summary.validation_objective_history["ndcg@32"],
            atol=1e-3,
        )



"""
1. Modify ``CMakeLists.txt`` in source tree and ``python-package/xgboost/VERSION`` if
needed, run CMake .
    If this is a RC release, the Python version has the form <major>.<minor>.<patch>rc1
2. Modify ``DESCRIPTION`` and ``configure.ac`` in R-package. Run ``autoreconf``.
3. Run ``mvn`` in ``jvm-packages``
    If this is a RC release, the version for JVM packages has the form
    <major>.<minor>.<patch>-RC1
"""

import argparse
import datetime
import os
import re
import subprocess
import sys
import tempfile

from pypi_variants import make_pyproject
from test_utils import JVM_PACKAGES, PY_PACKAGE, R_PACKAGE, ROOT, cd


@cd(ROOT)
def cmake(major: int, minor: int, patch: int) -> None:
    version = f"{major}.{minor}.{patch}"
    with open("CMakeLists.txt", "r") as fd:
        cmakelist = fd.read()
    pattern = r"project\(xgboost LANGUAGES .* VERSION ([0-9]+\.[0-9]+\.[0-9]+)\)"
    matched = re.search(pattern, cmakelist)
    assert matched, "Couldn't find the version string in CMakeLists.txt."
    print(matched.start(1), matched.end(1))
    cmakelist = cmakelist[: matched.start(1)] + version + cmakelist[matched.end(1) :]
    with open("CMakeLists.txt", "w") as fd:
        fd.write(cmakelist)

    with tempfile.TemporaryDirectory() as tmpdir:
        subprocess.call(["cmake", "-S", ".", "-B", tmpdir])


@cd(PY_PACKAGE)
def pypkg(
    major: int, minor: int, patch: int, rc: int, is_rc: bool, is_dev: bool
) -> None:
    version = f"{major}.{minor}.{patch}"
    pyver = version
    if is_rc:
        pyver = pyver + f"rc{rc}"
    if is_dev:
        pyver = pyver + "-dev"

    pyver_path = os.path.join("xgboost", "VERSION")
    with open(pyver_path, "w") as fd:
        fd.write(pyver + "\n")

    for pyprj_file in ["pyproject.toml.in", "pyproject.toml.stub.in"]:
        pyprj_path = os.path.join(pyprj_file)
        with open(pyprj_path, "r") as fd:
            pyprj = fd.read()
        matched = re.search('version = "' + r"([0-9]+\.[0-9]+\.[0-9]+.*)" + '"', pyprj)
        assert matched, "Couldn't find version string in pyproject.toml."
        pyprj = pyprj[: matched.start(1)] + pyver + pyprj[matched.end(1) :]
        with open(pyprj_path, "w") as fd:
            fd.write(pyprj)

    make_pyproject(use_suffix="na", require_nccl_dep="cu12")


@cd(R_PACKAGE)
def rpkg(major: int, minor: int, patch: int, is_dev: bool) -> None:
    if is_dev:
        version = f"{major}.{minor}.{patch}.0"
    else:
        version = f"{major}.{minor}.{patch}.1"
    # Version: 2.0.0.1
    desc_path = "DESCRIPTION"
    with open(desc_path, "r") as fd:
        description = fd.read()
        pattern = r"Version:\ ([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)"
        matched = re.search(pattern, description)
        assert matched, "Couldn't find version string in DESCRIPTION."
        description = (
            description[: matched.start(1)] + version + description[matched.end(1) :]
        )
        pattern = r"Date:\ ([0-9]+\-[0-9]+\-[0-9]+)"
        today = datetime.date.today()
        matched = re.search(pattern, description)
        assert matched, "Couldn't find date string in DESCRIPTION."
        description = (
            description[: matched.start(1)] + str(today) + description[matched.end(1) :]
        )
    with open(desc_path, "w") as fd:
        fd.write(description)

    config_path = "configure.ac"
    # AC_INIT([xgboost],[2.0.0],[],[xgboost],[])
    version = f"{major}.{minor}.{patch}"
    with open(config_path, "r") as fd:
        config = fd.read()
        pattern = (
            r"AC_INIT\(\[xgboost\],\[([0-9]+\.[0-9]+\.[0-9]+)\],\[\],\[xgboost\],\[\]\)"
        )
        matched = re.search(pattern, config)
        assert matched, "Couldn't find version string in configure.ac"
        config = config[: matched.start(1)] + version + config[matched.end(1) :]

    with open(config_path, "w") as fd:
        fd.write(config)

    subprocess.check_call(["autoreconf"])


@cd(JVM_PACKAGES)
def jvmpkgs(
    major: int, minor: int, patch: int, rc: int, is_rc: bool, is_dev: bool
) -> None:
    version = f"{major}.{minor}.{patch}"
    if is_dev:
        version += "-SNAPSHOT"
    if is_rc:
        version += f"-RC{rc}"
    subprocess.check_call(["mvn", "versions:set", f"-DnewVersion={version}"])


@cd(ROOT)
def main(args: argparse.Namespace) -> None:
    major = args.major
    minor = args.minor
    patch = args.patch
    rc = args.rc
    is_rc = args.is_rc
    is_dev = args.is_dev
    if is_rc and is_dev:
        raise ValueError("It cannot be both a rc and a dev branch.")
    if is_rc:
        assert rc >= 1, "RC version starts from 1."
    else:
        assert rc == 0, "RC is not used."

    cmake(major, minor, patch)
    pypkg(major, minor, patch, rc, is_rc, is_dev)
    rpkg(major, minor, patch, is_dev=is_dev)
    jvmpkgs(major, minor, patch, rc, is_rc, is_dev)

    print("""

Please examine the changes and commit. Be aware that mvn might leave backup files in the
source tree.

""")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--major", type=int, required=True)
    parser.add_argument("--minor", type=int, required=True)
    parser.add_argument("--patch", type=int, required=True)
    parser.add_argument("--rc", type=int, default=0)
    parser.add_argument("--is-rc", action="store_true")
    parser.add_argument("--is-dev", action="store_true")
    args = parser.parse_args()
    try:
        main(args)
    except Exception as e:
        print("Error:", e, file=sys.stderr)
        exit(-1)



from typing import Type

import numpy as np
import pytest

import xgboost as xgb
from xgboost import testing as tm
from xgboost.testing.monotone_constraints import training_dset, x, y

dpath = "demo/data/"


def is_increasing(y):
    return np.count_nonzero(np.diff(y) < 0.0) == 0


def is_decreasing(y):
    return np.count_nonzero(np.diff(y) > 0.0) == 0


def is_correctly_constrained(learner, feature_names=None):
    n = 100
    variable_x = np.linspace(0, 1, n).reshape((n, 1))
    fixed_xs_values = np.linspace(0, 1, n)

    for i in range(n):
        fixed_x = fixed_xs_values[i] * np.ones((n, 1))
        monotonically_increasing_x = np.column_stack((variable_x, fixed_x))
        monotonically_increasing_dset = xgb.DMatrix(
            monotonically_increasing_x, feature_names=feature_names
        )
        monotonically_increasing_y = learner.predict(monotonically_increasing_dset)

        monotonically_decreasing_x = np.column_stack((fixed_x, variable_x))
        monotonically_decreasing_dset = xgb.DMatrix(
            monotonically_decreasing_x, feature_names=feature_names
        )
        monotonically_decreasing_y = learner.predict(monotonically_decreasing_dset)

        if not (
            is_increasing(monotonically_increasing_y)
            and is_decreasing(monotonically_decreasing_y)
        ):
            return False

    return True


class TestMonotoneConstraints:
    def test_monotone_constraints_for_exact_tree_method(self) -> None:

        # first check monotonicity for the 'exact' tree method
        params_for_constrained_exact_method = {
            "tree_method": "exact",
            "verbosity": 1,
            "monotone_constraints": "(1, -1)",
        }
        constrained_exact_method = xgb.train(
            params_for_constrained_exact_method, training_dset
        )
        assert is_correctly_constrained(constrained_exact_method)

    @pytest.mark.parametrize(
        "tree_method,policy",
        [
            ("hist", "depthwise"),
            ("approx", "depthwise"),
            ("hist", "lossguide"),
            ("approx", "lossguide"),
        ],
    )
    def test_monotone_constraints(self, tree_method: str, policy: str) -> None:
        params_for_constrained = {
            "tree_method": tree_method,
            "grow_policy": policy,
            "monotone_constraints": "(1, -1)",
        }
        constrained = xgb.train(params_for_constrained, training_dset)
        assert is_correctly_constrained(constrained)

    def test_monotone_constraints_tuple(self) -> None:
        params_for_constrained = {"monotone_constraints": (1, -1)}
        constrained = xgb.train(params_for_constrained, training_dset)
        assert is_correctly_constrained(constrained)

    @pytest.mark.parametrize("format", [dict, list])
    def test_monotone_constraints_feature_names(self, format: Type) -> None:

        # next check monotonicity when initializing monotone_constraints by feature names
        params = {
            "tree_method": "hist",
            "grow_policy": "lossguide",
            "monotone_constraints": {"feature_0": 1, "feature_1": -1},
        }

        if format == list:
            params = list(params.items())

        with pytest.raises(ValueError):
            xgb.train(params, training_dset)

        feature_names = ["feature_0", "feature_2"]
        training_dset_w_feature_names = xgb.DMatrix(
            x, label=y, feature_names=feature_names
        )

        with pytest.raises(ValueError):
            xgb.train(params, training_dset_w_feature_names)

        feature_names = ["feature_0", "feature_1"]
        training_dset_w_feature_names = xgb.DMatrix(
            x, label=y, feature_names=feature_names
        )

        constrained_learner = xgb.train(params, training_dset_w_feature_names)

        assert is_correctly_constrained(constrained_learner, feature_names)

    @pytest.mark.skipif(**tm.no_sklearn())
    def test_training_accuracy(self) -> None:
        from sklearn.metrics import accuracy_score

        dtrain = xgb.DMatrix(dpath + "agaricus.txt.train?indexing_mode=1&format=libsvm")
        dtest = xgb.DMatrix(dpath + "agaricus.txt.test?indexing_mode=1&format=libsvm")
        params = {
            "eta": 1,
            "max_depth": 6,
            "objective": "binary:logistic",
            "tree_method": "hist",
            "monotone_constraints": "(1, 0)",
        }
        num_boost_round = 5

        params["grow_policy"] = "lossguide"
        bst = xgb.train(params, dtrain, num_boost_round)
        pred_dtest = bst.predict(dtest) < 0.5
        assert accuracy_score(dtest.get_label(), pred_dtest) < 0.1

        params["grow_policy"] = "depthwise"
        bst = xgb.train(params, dtrain, num_boost_round)
        pred_dtest = bst.predict(dtest) < 0.5
        assert accuracy_score(dtest.get_label(), pred_dtest) < 0.1



"""
Example of training survival model with Dask on CPU
===================================================

"""

import os

import dask.array as da
import dask.dataframe as dd
from dask.distributed import Client, LocalCluster

from xgboost import dask as dxgb
from xgboost.dask import DaskDMatrix


def main(client: Client) -> da.Array:
    # Load an example survival data from CSV into a Dask data frame.
    # The Veterans' Administration Lung Cancer Trial
    # The Statistical Analysis of Failure Time Data by Kalbfleisch J. and Prentice R (1980)
    CURRENT_DIR = os.path.dirname(__file__)
    df = dd.read_csv(
        os.path.join(CURRENT_DIR, os.pardir, "data", "veterans_lung_cancer.csv")
    )

    # DaskDMatrix acts like normal DMatrix, works as a proxy for local
    # DMatrix scatter around workers.
    # For AFT survival, you'd need to extract the lower and upper bounds for the label
    # and pass them as arguments to DaskDMatrix.
    y_lower_bound = df["Survival_label_lower_bound"]
    y_upper_bound = df["Survival_label_upper_bound"]
    X = df.drop(["Survival_label_lower_bound", "Survival_label_upper_bound"], axis=1)
    dtrain = DaskDMatrix(
        client, X, label_lower_bound=y_lower_bound, label_upper_bound=y_upper_bound
    )

    # Use train method from xgboost.dask instead of xgboost.  This
    # distributed version of train returns a dictionary containing the
    # resulting booster and evaluation history obtained from
    # evaluation metrics.
    params = {
        "verbosity": 1,
        "objective": "survival:aft",
        "eval_metric": "aft-nloglik",
        "learning_rate": 0.05,
        "aft_loss_distribution_scale": 1.20,
        "aft_loss_distribution": "normal",
        "max_depth": 6,
        "lambda": 0.01,
        "alpha": 0.02,
    }
    output = dxgb.train(
        client, params, dtrain, num_boost_round=100, evals=[(dtrain, "train")]
    )
    bst = output["booster"]
    history = output["history"]

    # you can pass output directly into `predict` too.
    prediction = dxgb.predict(client, bst, dtrain)
    print("Evaluation history: ", history)

    # Uncomment the following line to save the model to the disk
    # bst.save_model('survival_model.json')

    return prediction


if __name__ == "__main__":
    # or use other clusters for scaling
    with LocalCluster(n_workers=7, threads_per_worker=4) as cluster:
        with Client(cluster) as client:
            main(client)



"""
Demo for accessing the xgboost eval metrics by using sklearn interface
======================================================================
"""

import numpy as np
from sklearn.datasets import make_hastie_10_2

import xgboost as xgb

X, y = make_hastie_10_2(n_samples=2000, random_state=42)

# Map labels from {-1, 1} to {0, 1}
labels, y = np.unique(y, return_inverse=True)

X_train, X_test = X[:1600], X[1600:]
y_train, y_test = y[:1600], y[1600:]

param_dist = {"objective": "binary:logistic", "n_estimators": 2}

clf = xgb.XGBModel(
    **param_dist,
    eval_metric="logloss",
)
# Or you can use: clf = xgb.XGBClassifier(**param_dist)

clf.fit(
    X_train,
    y_train,
    eval_set=[(X_train, y_train), (X_test, y_test)],
    verbose=True,
)

# Load evals result by calling the evals_result() function
evals_result = clf.evals_result()

print("Access logloss metric directly from validation_0:")
print(evals_result["validation_0"]["logloss"])

print("")
print("Access metrics through a loop:")
for e_name, e_mtrs in evals_result.items():
    print("- {}".format(e_name))
    for e_mtr_name, e_mtr_vals in e_mtrs.items():
        print("   - {}".format(e_mtr_name))
        print("      - {}".format(e_mtr_vals))

print("")
print("Access complete dict:")
print(evals_result)



"""
Demo for prediction using number of trees
=========================================
"""
import os

import numpy as np
from sklearn.datasets import load_svmlight_file

import xgboost as xgb

CURRENT_DIR = os.path.dirname(__file__)
train = os.path.join(CURRENT_DIR, "../data/agaricus.txt.train")
test = os.path.join(CURRENT_DIR, "../data/agaricus.txt.test")


def native_interface() -> None:
    # load data in do training
    dtrain = xgb.DMatrix(train + "?format=libsvm")
    dtest = xgb.DMatrix(test + "?format=libsvm")
    param = {"max_depth": 2, "eta": 1, "objective": "binary:logistic"}
    watchlist = [(dtest, "eval"), (dtrain, "train")]
    num_round = 3
    bst = xgb.train(param, dtrain, num_round, watchlist)

    print("start testing prediction from first n trees")
    # predict using first 1 tree
    label = dtest.get_label()
    ypred1 = bst.predict(dtest, iteration_range=(0, 1))
    # by default, we predict using all the trees
    ypred2 = bst.predict(dtest)

    print("error of ypred1=%f" % (np.sum((ypred1 > 0.5) != label) / float(len(label))))
    print("error of ypred2=%f" % (np.sum((ypred2 > 0.5) != label) / float(len(label))))


def sklearn_interface() -> None:
    X_train, y_train = load_svmlight_file(train)
    X_test, y_test = load_svmlight_file(test)
    clf = xgb.XGBClassifier(n_estimators=3, max_depth=2, eta=1)
    clf.fit(X_train, y_train, eval_set=[(X_test, y_test)])
    assert clf.n_classes_ == 2

    print("start testing prediction from first n trees")
    # predict using first 1 tree
    ypred1 = clf.predict(X_test, iteration_range=(0, 1))
    # by default, we predict using all the trees
    ypred2 = clf.predict(X_test)

    print(
        "error of ypred1=%f" % (np.sum((ypred1 > 0.5) != y_test) / float(len(y_test)))
    )
    print(
        "error of ypred2=%f" % (np.sum((ypred2 > 0.5) != y_test) / float(len(y_test)))
    )


if __name__ == "__main__":
    native_interface()
    sklearn_interface()



import os
import subprocess
import sys
from typing import List

from test_utils import PY_PACKAGE, ROOT, cd, print_time, record_time


class TypeCheckPaths:
    """The paths mypy runs on."""

    MYPY = (
        # core
        "python-package/",
        # tests
        "tests/python/generate_models.py",
        "tests/python/test_model_compatibility.py",
        "tests/python/test_collective.py",
        "tests/python/test_demos.py",
        "tests/python/test_data_iterator.py",
        "tests/python/test_multi_target.py",
        "tests/python/test_intercept.py",
        "tests/python/test_model_io.py",
        "tests/python/test_ordinal.py",
        "tests/python/test_interaction_constraints.py",
        "tests/python-gpu/",
        "tests/test_distributed/test_federated/",
        "tests/test_distributed/test_gpu_federated/",
        "tests/test_distributed/test_with_dask/",
        "tests/test_distributed/test_with_spark/test_data.py",
        "tests/test_distributed/test_gpu_with_dask/",
        # demo
        "demo/dask/",
        "demo/guide-python/",
        "demo/aft_survival/aft_survival_viz_demo.py",
        # CI
        "ops/",
    )


def check_cmd_print_failure_assistance(cmd: List[str]) -> bool:
    if subprocess.run(cmd).returncode == 0:
        return True

    subprocess.run([cmd[0], "--version"])
    msg = """
Please run the following command on your machine to address the error:

    """
    msg += " ".join(cmd)
    print(msg, file=sys.stderr)
    return False


@record_time
@cd(PY_PACKAGE)
def run_mypy(rel_path: str) -> bool:
    cmd = ["mypy", os.path.join(ROOT, rel_path)]
    return check_cmd_print_failure_assistance(cmd)


@record_time
def main() -> None:
    mypy_results = [run_mypy(path) for path in TypeCheckPaths.MYPY]
    if not all(mypy_results):
        sys.exit(-1)


if __name__ == "__main__":
    try:
        main()
    finally:
        print_time()



# pylint: disable=unbalanced-tuple-unpacking, too-many-locals
"""Tests for federated learning."""

import multiprocessing
import os
import subprocess
import tempfile
import time
from typing import List, cast

from sklearn.datasets import dump_svmlight_file, load_svmlight_file
from sklearn.model_selection import train_test_split

import xgboost as xgb
import xgboost.federated
from xgboost import testing as tm

from .._typing import EvalsLog
from ..collective import _Args as CollArgs

SERVER_KEY = "server-key.pem"
SERVER_CERT = "server-cert.pem"
CLIENT_KEY = "client-key.pem"
CLIENT_CERT = "client-cert.pem"


def run_server(port: int, world_size: int, with_ssl: bool) -> None:
    """Run federated server for test."""
    if with_ssl:
        xgboost.federated.run_federated_server(
            world_size,
            port,
            server_key_path=SERVER_KEY,
            server_cert_path=SERVER_CERT,
            client_cert_path=CLIENT_CERT,
        )
    else:
        xgboost.federated.run_federated_server(world_size, port)


def run_worker(
    port: int, world_size: int, rank: int, with_ssl: bool, device: str
) -> None:
    """Run federated client worker for test."""
    comm_env: CollArgs = {
        "dmlc_communicator": "federated",
        "federated_server_address": f"localhost:{port}",
        "federated_world_size": world_size,
        "federated_rank": rank,
    }
    if with_ssl:
        comm_env["federated_server_cert_path"] = SERVER_CERT
        comm_env["federated_client_key_path"] = CLIENT_KEY
        comm_env["federated_client_cert_path"] = CLIENT_CERT

    cpu_count = os.cpu_count()
    assert cpu_count is not None
    n_threads = cpu_count // world_size

    # Always call this before using distributed module
    with xgb.collective.CommunicatorContext(**comm_env):
        # Load file, file will not be sharded in federated mode.
        X, y = load_svmlight_file(f"agaricus.txt-{rank}.train")
        dtrain = xgb.DMatrix(X, y)
        X, y = load_svmlight_file(f"agaricus.txt-{rank}.test")
        dtest = xgb.DMatrix(X, y)

        # Specify parameters via map, definition are same as c++ version
        param = {
            "max_depth": 2,
            "eta": 1,
            "objective": "binary:logistic",
            "nthread": n_threads,
            "tree_method": "hist",
            "device": device,
        }

        # Specify validations set to watch performance
        watchlist = [(dtest, "eval"), (dtrain, "train")]
        num_round = 20

        # Run training, all the features in training API is available.
        results: EvalsLog = {}
        bst = xgb.train(
            param,
            dtrain,
            num_round,
            evals=watchlist,
            early_stopping_rounds=2,
            evals_result=results,
        )
        assert tm.non_increasing(cast(List[float], results["train"]["logloss"]))
        assert tm.non_increasing(cast(List[float], results["eval"]["logloss"]))

        # save the model, only ask process 0 to save the model.
        if xgb.collective.get_rank() == 0:
            with tempfile.TemporaryDirectory() as tmpdir:
                bst.save_model(os.path.join(tmpdir, "model.json"))
            xgb.collective.communicator_print("Finished training\n")


def run_federated(world_size: int, with_ssl: bool, use_gpu: bool) -> None:
    """Launcher for clients and the server."""
    port = 9091

    server = multiprocessing.Process(
        target=run_server, args=(port, world_size, with_ssl)
    )
    server.start()
    time.sleep(1)
    if not server.is_alive():
        raise ValueError("Error starting Federated Learning server")

    workers = []
    for rank in range(world_size):
        device = f"cuda:{rank}" if use_gpu else "cpu"
        worker = multiprocessing.Process(
            target=run_worker, args=(port, world_size, rank, with_ssl, device)
        )
        workers.append(worker)
        worker.start()
    for worker in workers:
        worker.join()
    server.terminate()


def run_federated_learning(with_ssl: bool, use_gpu: bool, test_path: str) -> None:
    """Run federated learning tests."""
    n_workers = 2

    if with_ssl:
        command = "openssl req -x509 -newkey rsa:2048 -days 7 -nodes -keyout {part}-key.pem -out {part}-cert.pem -subj /C=US/CN=localhost"  # pylint: disable=line-too-long
        server_key = command.format(part="server").split()
        subprocess.check_call(server_key)
        client_key = command.format(part="client").split()
        subprocess.check_call(client_key)

    train_path = os.path.join(tm.data_dir(test_path), "agaricus.txt.train")
    test_path = os.path.join(tm.data_dir(test_path), "agaricus.txt.test")

    X_train, y_train = load_svmlight_file(train_path)
    X_test, y_test = load_svmlight_file(test_path)

    X0, X1, y0, y1 = train_test_split(X_train, y_train, test_size=0.5)
    X0_valid, X1_valid, y0_valid, y1_valid = train_test_split(
        X_test, y_test, test_size=0.5
    )

    dump_svmlight_file(X0, y0, "agaricus.txt-0.train")
    dump_svmlight_file(X0_valid, y0_valid, "agaricus.txt-0.test")

    dump_svmlight_file(X1, y1, "agaricus.txt-1.train")
    dump_svmlight_file(X1_valid, y1_valid, "agaricus.txt-1.test")

    run_federated(world_size=n_workers, with_ssl=with_ssl, use_gpu=use_gpu)



from typing import Tuple

import numpy as np
import pytest

import xgboost as xgb
from xgboost import testing as tm
from xgboost.testing.updater import get_basescore

rng = np.random.RandomState(1994)


class TestEarlyStopping:
    @pytest.mark.skipif(**tm.no_sklearn())
    def test_early_stopping_nonparallel(self):
        from sklearn.datasets import load_digits
        from sklearn.model_selection import train_test_split

        X, y = load_digits(n_class=2, return_X_y=True)
        X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)
        clf1 = xgb.XGBClassifier(
            learning_rate=0.1, early_stopping_rounds=5, eval_metric="auc"
        )
        clf1.fit(X_train, y_train, eval_set=[(X_test, y_test)])
        clf2 = xgb.XGBClassifier(
            learning_rate=0.1, early_stopping_rounds=4, eval_metric="auc"
        )
        clf2.fit(X_train, y_train, eval_set=[(X_test, y_test)])
        # should be the same
        assert clf1.best_score == clf2.best_score
        assert clf1.best_score != 1
        # check overfit
        clf3 = xgb.XGBClassifier(
            learning_rate=0.1, eval_metric="auc", early_stopping_rounds=10
        )
        clf3.fit(X_train, y_train, eval_set=[(X_test, y_test)])
        base_score = get_basescore(clf3)
        assert 0.53 > base_score[0] > 0.5

        clf3 = xgb.XGBClassifier(
            learning_rate=0.1,
            base_score=0.5,
            eval_metric="auc",
            early_stopping_rounds=10,
        )
        clf3.fit(X_train, y_train, eval_set=[(X_test, y_test)])

        assert clf3.best_score == 1

    @staticmethod
    def assert_metrics_length(cv, expected_length):
        for key, value in cv.items():
            assert len(value) == expected_length

    @pytest.mark.skipif(**tm.no_sklearn())
    def test_cv_early_stopping(self) -> None:
        from sklearn.datasets import load_digits

        X, y = load_digits(n_class=2, return_X_y=True)
        dm = xgb.DMatrix(X, label=y)
        params = {
            "max_depth": 2,
            "eta": 1,
            "objective": "binary:logistic",
            "eval_metric": "error",
        }

        def evalerror(preds: np.ndarray, dtrain: xgb.DMatrix) -> Tuple[str, float]:
            from sklearn.metrics import mean_squared_error

            labels = dtrain.get_label()
            return "rmse", mean_squared_error(labels, preds)

        cv = xgb.cv(params, dm, num_boost_round=10, nfold=10, early_stopping_rounds=10)
        self.assert_metrics_length(cv, 10)
        cv = xgb.cv(params, dm, num_boost_round=10, nfold=10, early_stopping_rounds=5)
        self.assert_metrics_length(cv, 3)
        cv = xgb.cv(params, dm, num_boost_round=10, nfold=10, early_stopping_rounds=1)
        self.assert_metrics_length(cv, 1)

        cv = xgb.cv(
            params,
            dm,
            num_boost_round=10,
            nfold=10,
            custom_metric=evalerror,
            early_stopping_rounds=10,
        )
        self.assert_metrics_length(cv, 10)
        cv = xgb.cv(
            params,
            dm,
            num_boost_round=10,
            nfold=10,
            custom_metric=evalerror,
            early_stopping_rounds=1,
        )
        self.assert_metrics_length(cv, 5)
        cv = xgb.cv(
            params,
            dm,
            num_boost_round=10,
            nfold=10,
            custom_metric=evalerror,
            maximize=True,
            early_stopping_rounds=1,
        )
        self.assert_metrics_length(cv, 1)

        with pytest.raises(ValueError, match="`save_best`"):
            cv = xgb.cv(
                params,
                dm,
                num_boost_round=10,
                nfold=10,
                early_stopping_rounds=1,
                callbacks=[xgb.callback.EarlyStopping(3, save_best=True)],
            )

    @pytest.mark.skipif(**tm.no_sklearn())
    @pytest.mark.skipif(**tm.no_pandas())
    def test_cv_early_stopping_with_multiple_eval_sets_and_metrics(self):
        from sklearn.datasets import load_breast_cancer

        X, y = load_breast_cancer(return_X_y=True)
        dm = xgb.DMatrix(X, label=y)
        params = {"objective": "binary:logistic"}

        metrics = [
            ["auc"],
            ["error"],
            ["logloss"],
            ["logloss", "auc"],
            ["logloss", "error"],
            ["error", "logloss"],
        ]

        num_iteration_history = []

        # If more than one metrics is given, early stopping should use the last metric
        for i, m in enumerate(metrics):
            result = xgb.cv(
                params,
                dm,
                num_boost_round=1000,
                nfold=5,
                stratified=True,
                metrics=m,
                early_stopping_rounds=20,
                seed=42,
            )
            num_iteration_history.append(len(result))
            df = result["test-{}-mean".format(m[-1])]
            # When early stopping is invoked, the last metric should be as best it can be.
            if m[-1] == "auc":
                assert np.all(df <= df.iloc[-1])
            else:
                assert np.all(df >= df.iloc[-1])
        assert num_iteration_history[:3] == num_iteration_history[3:]



import numpy as np
import pytest

import xgboost as xgb
from xgboost import testing as tm
from xgboost.testing.plotting import run_categorical

try:
    import matplotlib

    matplotlib.use("Agg")
    from graphviz import Source
    from matplotlib.axes import Axes
except ImportError:
    pass

pytestmark = pytest.mark.skipif(**tm.no_multiple(tm.no_matplotlib(), tm.no_graphviz()))


class TestPlotting:
    def test_plotting(self):
        m, _ = tm.load_agaricus(__file__)
        booster = xgb.train(
            {"max_depth": 2, "eta": 1, "objective": "binary:logistic"},
            m,
            num_boost_round=2,
        )

        ax = xgb.plot_importance(booster)
        assert isinstance(ax, Axes)
        assert ax.get_title() == "Feature importance"
        assert ax.get_xlabel() == "Importance score"
        assert ax.get_ylabel() == "Features"
        assert len(ax.patches) == 4

        ax = xgb.plot_importance(booster, color="r", title="t", xlabel="x", ylabel="y")
        assert isinstance(ax, Axes)
        assert ax.get_title() == "t"
        assert ax.get_xlabel() == "x"
        assert ax.get_ylabel() == "y"
        assert len(ax.patches) == 4
        for p in ax.patches:
            assert p.get_facecolor() == (1.0, 0, 0, 1.0)  # red

        ax = xgb.plot_importance(
            booster, color=["r", "r", "b", "b"], title=None, xlabel=None, ylabel=None
        )
        assert isinstance(ax, Axes)
        assert ax.get_title() == ""
        assert ax.get_xlabel() == ""
        assert ax.get_ylabel() == ""
        assert len(ax.patches) == 4
        assert ax.patches[0].get_facecolor() == (1.0, 0, 0, 1.0)  # red
        assert ax.patches[1].get_facecolor() == (1.0, 0, 0, 1.0)  # red
        assert ax.patches[2].get_facecolor() == (0, 0, 1.0, 1.0)  # blue
        assert ax.patches[3].get_facecolor() == (0, 0, 1.0, 1.0)  # blue

        g = xgb.to_graphviz(booster, tree_idx=0)
        assert isinstance(g, Source)

        ax = xgb.plot_tree(booster, tree_idx=0)
        assert isinstance(ax, Axes)

    def test_importance_plot_lim(self):
        np.random.seed(1)
        dm = xgb.DMatrix(np.random.randn(100, 100), label=[0, 1] * 50)
        bst = xgb.train({}, dm)
        assert len(bst.get_fscore()) == 71
        ax = xgb.plot_importance(bst)
        assert ax.get_xlim() == (0.0, 11.0)
        assert ax.get_ylim() == (-1.0, 71.0)

        ax = xgb.plot_importance(bst, xlim=(0, 5), ylim=(10, 71))
        assert ax.get_xlim() == (0.0, 5.0)
        assert ax.get_ylim() == (10.0, 71.0)

    @pytest.mark.skipif(**tm.no_pandas())
    def test_categorical(self) -> None:
        run_categorical("approx", "cpu")



"""
Demo for prediction using individual trees and model slices
===========================================================
"""

import os

import numpy as np
from scipy.special import logit
from sklearn.datasets import load_svmlight_file

import xgboost as xgb

CURRENT_DIR = os.path.dirname(__file__)
train = os.path.join(CURRENT_DIR, "../data/agaricus.txt.train")
test = os.path.join(CURRENT_DIR, "../data/agaricus.txt.test")


def individual_tree() -> None:
    """Get prediction from each individual tree and combine them together."""
    X_train, y_train = load_svmlight_file(train)
    X_test, y_test = load_svmlight_file(test)
    Xy_train = xgb.QuantileDMatrix(X_train, y_train)

    n_rounds = 4
    # Specify the base score, otherwise xgboost will estimate one from the training
    # data.
    base_score = 0.5
    params = {
        "max_depth": 2,
        "eta": 1,
        "objective": "reg:logistic",
        "tree_method": "hist",
        "base_score": base_score,
    }
    booster = xgb.train(params, Xy_train, num_boost_round=n_rounds)

    # Use logit to inverse the base score back to raw leaf value (margin)
    scores = np.full((X_test.shape[0],), logit(base_score))
    for i in range(n_rounds):
        # - Use output_margin to get raw leaf values
        # - Use iteration_range to get prediction for only one tree
        # - Use previous prediction as base marign for the model
        Xy_test = xgb.DMatrix(X_test, base_margin=scores)

        if i == n_rounds - 1:
            # last round, get the transformed prediction
            scores = booster.predict(
                Xy_test, iteration_range=(i, i + 1), output_margin=False
            )
        else:
            # get raw leaf value for accumulation
            scores = booster.predict(
                Xy_test, iteration_range=(i, i + 1), output_margin=True
            )

    full = booster.predict(xgb.DMatrix(X_test), output_margin=False)
    np.testing.assert_allclose(scores, full)


def model_slices() -> None:
    """Inference with each individual tree using model slices."""
    X_train, y_train = load_svmlight_file(train)
    X_test, y_test = load_svmlight_file(test)
    Xy_train = xgb.QuantileDMatrix(X_train, y_train)

    n_rounds = 4
    # Specify the base score, otherwise xgboost will estimate one from the training
    # data.
    base_score = 0.5
    params = {
        "max_depth": 2,
        "eta": 1,
        "objective": "reg:logistic",
        "tree_method": "hist",
        "base_score": base_score,
    }
    booster = xgb.train(params, Xy_train, num_boost_round=n_rounds)
    trees = [booster[t] for t in range(n_rounds)]

    # Use logit to inverse the base score back to raw leaf value (margin)
    scores = np.full((X_test.shape[0],), logit(base_score))
    for i, t in enumerate(trees):
        # Feed previous scores into base margin.
        Xy_test = xgb.DMatrix(X_test, base_margin=scores)

        if i == n_rounds - 1:
            # last round, get the transformed prediction
            scores = t.predict(Xy_test, output_margin=False)
        else:
            # get raw leaf value for accumulation
            scores = t.predict(Xy_test, output_margin=True)

    full = booster.predict(xgb.DMatrix(X_test), output_margin=False)
    np.testing.assert_allclose(scores, full)


if __name__ == "__main__":
    individual_tree()
    model_slices()



from typing import List

import numpy as np
import pandas as pd
import pytest
from xgboost import testing as tm

pytestmark = [pytest.mark.skipif(**tm.no_spark())]

from xgboost import DMatrix, QuantileDMatrix
from xgboost.spark.data import (
    _read_csr_matrix_from_unwrapped_spark_vec,
    alias,
    create_dmatrix_from_partitions,
    stack_series,
)


def test_stack() -> None:
    a = pd.DataFrame({"a": [[1, 2], [3, 4]]})
    b = stack_series(a["a"])
    assert b.shape == (2, 2)

    a = pd.DataFrame({"a": [[1], [3]]})
    b = stack_series(a["a"])
    assert b.shape == (2, 1)

    a = pd.DataFrame({"a": [np.array([1, 2]), np.array([3, 4])]})
    b = stack_series(a["a"])
    assert b.shape == (2, 2)

    a = pd.DataFrame({"a": [np.array([1]), np.array([3])]})
    b = stack_series(a["a"])
    assert b.shape == (2, 1)


def run_dmatrix_ctor(is_feature_cols: bool, is_qdm: bool, on_gpu: bool) -> None:
    rng = np.random.default_rng(0)
    dfs: List[pd.DataFrame] = []
    n_features = 16
    n_samples_per_batch = 16
    n_batches = 10
    feature_types = ["float"] * n_features

    for i in range(n_batches):
        X = rng.normal(loc=0, size=256).reshape(n_samples_per_batch, n_features)
        y = rng.normal(loc=0, size=n_samples_per_batch)
        m = rng.normal(loc=0, size=n_samples_per_batch)
        w = rng.normal(loc=0.5, scale=0.5, size=n_samples_per_batch)
        w -= w.min()

        valid = rng.binomial(n=1, p=0.5, size=16).astype(np.bool_)

        df = pd.DataFrame(
            {alias.label: y, alias.margin: m, alias.weight: w, alias.valid: valid}
        )
        if is_feature_cols:
            for j in range(X.shape[1]):
                df[f"feat-{j}"] = pd.Series(X[:, j])
        else:
            df[alias.data] = pd.Series(list(X))
        dfs.append(df)

    kwargs = {"feature_types": feature_types}
    device_id = 0 if on_gpu else None
    cols = [f"feat-{i}" for i in range(n_features)]
    feature_cols = cols if is_feature_cols else None
    train_Xy, valid_Xy = create_dmatrix_from_partitions(
        iterator=iter(dfs),
        feature_cols=feature_cols,
        dev_ordinal=device_id,
        use_qdm=is_qdm,
        kwargs=kwargs,
        enable_sparse_data_optim=False,
        has_validation_col=True,
    )

    if is_qdm:
        assert isinstance(train_Xy, QuantileDMatrix)
        assert isinstance(valid_Xy, QuantileDMatrix)
    else:
        assert not isinstance(train_Xy, QuantileDMatrix)
        assert isinstance(train_Xy, DMatrix)
        assert not isinstance(valid_Xy, QuantileDMatrix)
        assert isinstance(valid_Xy, DMatrix)

    assert valid_Xy is not None
    assert valid_Xy.num_row() + train_Xy.num_row() == n_samples_per_batch * n_batches
    assert train_Xy.num_col() == n_features
    assert valid_Xy.num_col() == n_features

    df = pd.concat(dfs, axis=0)
    df_train = df.loc[~df[alias.valid], :]
    df_valid = df.loc[df[alias.valid], :]

    assert df_train.shape[0] == train_Xy.num_row()
    assert df_valid.shape[0] == valid_Xy.num_row()

    # margin
    np.testing.assert_allclose(
        df_train[alias.margin].to_numpy(), train_Xy.get_base_margin()
    )
    np.testing.assert_allclose(
        df_valid[alias.margin].to_numpy(), valid_Xy.get_base_margin()
    )
    # weight
    np.testing.assert_allclose(df_train[alias.weight].to_numpy(), train_Xy.get_weight())
    np.testing.assert_allclose(df_valid[alias.weight].to_numpy(), valid_Xy.get_weight())
    # label
    np.testing.assert_allclose(df_train[alias.label].to_numpy(), train_Xy.get_label())
    np.testing.assert_allclose(df_valid[alias.label].to_numpy(), valid_Xy.get_label())

    np.testing.assert_equal(train_Xy.feature_types, feature_types)
    np.testing.assert_equal(valid_Xy.feature_types, feature_types)


@pytest.mark.parametrize(
    "is_feature_cols,is_qdm",
    [(True, True), (True, False), (False, True), (False, False)],
)
def test_dmatrix_ctor(is_feature_cols: bool, is_qdm: bool) -> None:
    run_dmatrix_ctor(is_feature_cols, is_qdm, on_gpu=False)


@pytest.mark.skipif(**tm.no_cudf())
@tm.timeout(120)
@pytest.mark.parametrize(
    "is_feature_cols,is_qdm",
    [(True, True), (True, False), (False, True), (False, False)],
)
def test_dmatrix_ctor_gpu(is_feature_cols: bool, is_qdm: bool) -> None:
    run_dmatrix_ctor(is_feature_cols, is_qdm, on_gpu=True)


def test_read_csr_matrix_from_unwrapped_spark_vec() -> None:
    from scipy.sparse import csr_matrix

    pd1 = pd.DataFrame(
        {
            "featureVectorType": [0, 1, 1, 0],
            "featureVectorSize": [3, None, None, 3],
            "featureVectorIndices": [
                np.array([0, 2], dtype=np.int32),
                None,
                None,
                np.array([1, 2], dtype=np.int32),
            ],
            "featureVectorValues": [
                np.array([3.0, 0.0], dtype=np.float64),
                np.array([13.0, 14.0, 0.0], dtype=np.float64),
                np.array([0.0, 24.0, 25.0], dtype=np.float64),
                np.array([0.0, 35.0], dtype=np.float64),
            ],
        }
    )
    sm = _read_csr_matrix_from_unwrapped_spark_vec(pd1)
    assert isinstance(sm, csr_matrix)

    np.testing.assert_array_equal(
        sm.data, [3.0, 0.0, 13.0, 14.0, 0.0, 0.0, 24.0, 25.0, 0.0, 35.0]
    )
    np.testing.assert_array_equal(sm.indptr, [0, 2, 5, 8, 10])
    np.testing.assert_array_equal(sm.indices, [0, 2, 0, 1, 2, 0, 1, 2, 1, 2])
    assert sm.shape == (4, 3)



"""
Getting started with XGBoost
============================

This is a simple example of using the native XGBoost interface, there are other
interfaces in the Python package like scikit-learn interface and Dask interface.


See :doc:`/python/python_intro` and :doc:`/tutorials/index` for other references.

"""
import os
import pickle

import numpy as np
from sklearn.datasets import load_svmlight_file

import xgboost as xgb

# Make sure the demo knows where to load the data.
CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
XGBOOST_ROOT_DIR = os.path.dirname(os.path.dirname(CURRENT_DIR))
DEMO_DIR = os.path.join(XGBOOST_ROOT_DIR, "demo")

# X is a scipy csr matrix, XGBoost supports many other input types,
X, y = load_svmlight_file(os.path.join(DEMO_DIR, "data", "agaricus.txt.train"))
dtrain = xgb.DMatrix(X, y)
# validation set
X_test, y_test = load_svmlight_file(os.path.join(DEMO_DIR, "data", "agaricus.txt.test"))
dtest = xgb.DMatrix(X_test, y_test)

# specify parameters via map, definition are same as c++ version
param = {"max_depth": 2, "eta": 1, "objective": "binary:logistic"}

# specify validations set to watch performance
watchlist = [(dtest, "eval"), (dtrain, "train")]
# number of boosting rounds
num_round = 2
bst = xgb.train(param, dtrain, num_boost_round=num_round, evals=watchlist)

# run prediction
preds = bst.predict(dtest)
labels = dtest.get_label()
print(
    "error=%f"
    % (
        sum(1 for i in range(len(preds)) if int(preds[i] > 0.5) != labels[i])
        / float(len(preds))
    )
)
bst.save_model("model-0.json")
# dump model
bst.dump_model("dump.raw.txt")
# dump model with feature map
bst.dump_model("dump.nice.txt", os.path.join(DEMO_DIR, "data/featmap.txt"))

# save dmatrix into binary buffer
dtest.save_binary("dtest.dmatrix")
# save model
bst.save_model("model-1.json")
# load model and data in
bst2 = xgb.Booster(model_file="model-1.json")
dtest2 = xgb.DMatrix("dtest.dmatrix")
preds2 = bst2.predict(dtest2)
# assert they are the same
assert np.sum(np.abs(preds2 - preds)) == 0

# alternatively, you can pickle the booster
pks = pickle.dumps(bst2)
# load model and data in
bst3 = pickle.loads(pks)
preds3 = bst3.predict(dtest2)
# assert they are the same
assert np.sum(np.abs(preds3 - preds)) == 0



"""Tests for running inplace prediction."""

from concurrent.futures import ThreadPoolExecutor
from typing import List, Type, Union

import numpy as np
import pandas as pd
import pytest
import xgboost as xgb
from scipy import sparse
from xgboost import testing as tm
from xgboost.testing.data import get_california_housing, np_dtypes, pd_dtypes
from xgboost.testing.predict import run_base_margin_vs_base_score, run_predict_leaf


def run_threaded_predict(X, rows, predict_func):
    results = []
    per_thread = 20
    with ThreadPoolExecutor(max_workers=10) as e:
        for i in range(0, rows, int(rows / per_thread)):
            if hasattr(X, "iloc"):
                predictor = X.iloc[i : i + per_thread, :]
            else:
                predictor = X[i : i + per_thread, ...]
            f = e.submit(predict_func, predictor)
            results.append(f)

    for f in results:
        assert f.result()


@pytest.mark.parametrize("DMatrixT", [xgb.DMatrix, xgb.QuantileDMatrix])
def test_predict_leaf(DMatrixT: Type[xgb.DMatrix]) -> None:
    run_predict_leaf("cpu", DMatrixT)


def test_predict_shape():
    X, y = get_california_housing()
    reg = xgb.XGBRegressor(n_estimators=1)
    reg.fit(X, y)
    predt = reg.get_booster().predict(xgb.DMatrix(X), strict_shape=True)
    assert len(predt.shape) == 2
    assert predt.shape[0] == X.shape[0]
    assert predt.shape[1] == 1

    contrib = reg.get_booster().predict(
        xgb.DMatrix(X), pred_contribs=True, strict_shape=True
    )
    assert len(contrib.shape) == 3
    assert contrib.shape[1] == 1

    contrib = reg.get_booster().predict(
        xgb.DMatrix(X), pred_contribs=True, approx_contribs=True
    )
    assert len(contrib.shape) == 2
    assert contrib.shape[1] == X.shape[1] + 1

    interaction = reg.get_booster().predict(
        xgb.DMatrix(X), pred_interactions=True, approx_contribs=True
    )
    assert len(interaction.shape) == 3
    assert interaction.shape[1] == X.shape[1] + 1
    assert interaction.shape[2] == X.shape[1] + 1

    interaction = reg.get_booster().predict(
        xgb.DMatrix(X), pred_interactions=True, approx_contribs=True, strict_shape=True
    )
    assert len(interaction.shape) == 4
    assert interaction.shape[1] == 1
    assert interaction.shape[2] == X.shape[1] + 1
    assert interaction.shape[3] == X.shape[1] + 1


def test_base_margin_vs_base_score() -> None:
    run_base_margin_vs_base_score("cpu")


class TestInplacePredict:
    """Tests for running inplace prediction"""

    @classmethod
    def setup_class(cls):
        cls.rows = 1000
        cls.cols = 10

        cls.missing = 11  # set to integer for testing

        cls.rng = np.random.RandomState(1994)

        cls.X = cls.rng.randn(cls.rows, cls.cols)
        missing_idx = [i for i in range(0, cls.cols, 4)]
        cls.X[:, missing_idx] = cls.missing  # set to be missing

        cls.y = cls.rng.randn(cls.rows)

        dtrain = xgb.DMatrix(cls.X, cls.y)
        cls.test = xgb.DMatrix(cls.X[:10, ...], missing=cls.missing)

        cls.num_boost_round = 10
        cls.booster = xgb.train({"tree_method": "hist"}, dtrain, num_boost_round=10)

    def test_predict(self):
        booster = self.booster
        X = self.X
        test = self.test

        predt_from_array = booster.inplace_predict(X[:10, ...], missing=self.missing)
        predt_from_dmatrix = booster.predict(test)

        X_obj = X.copy().astype(object)

        assert X_obj.dtype.hasobject is True
        assert X.dtype.hasobject is False
        np.testing.assert_allclose(
            booster.inplace_predict(X_obj), booster.inplace_predict(X)
        )

        np.testing.assert_allclose(predt_from_dmatrix, predt_from_array)

        predt_from_array = booster.inplace_predict(
            X[:10, ...], iteration_range=(0, 4), missing=self.missing
        )
        predt_from_dmatrix = booster.predict(test, iteration_range=(0, 4))

        np.testing.assert_allclose(predt_from_dmatrix, predt_from_array)

        with pytest.raises(ValueError):
            booster.predict(test, iteration_range=(0, booster.num_boosted_rounds() + 2))

        default = booster.predict(test)

        range_full = booster.predict(test, iteration_range=(0, self.num_boost_round))
        np.testing.assert_allclose(range_full, default)

        range_full = booster.predict(
            test, iteration_range=(0, booster.num_boosted_rounds())
        )
        np.testing.assert_allclose(range_full, default)

        def predict_dense(x):
            inplace_predt = booster.inplace_predict(x)
            d = xgb.DMatrix(x)
            copied_predt = booster.predict(d)
            return np.all(copied_predt == inplace_predt)

        for i in range(10):
            run_threaded_predict(X, self.rows, predict_dense)

        def predict_csr(x):
            inplace_predt = booster.inplace_predict(sparse.csr_matrix(x))
            d = xgb.DMatrix(x)
            copied_predt = booster.predict(d)
            return np.all(copied_predt == inplace_predt)

        for i in range(10):
            run_threaded_predict(X, self.rows, predict_csr)

    @pytest.mark.skipif(**tm.no_pandas())
    def test_predict_pd(self):
        X = self.X
        # construct it in column major style
        df = pd.DataFrame({str(i): X[:, i] for i in range(X.shape[1])})
        booster = self.booster
        df_predt = booster.inplace_predict(df)
        arr_predt = booster.inplace_predict(X)
        dmat_predt = booster.predict(xgb.DMatrix(X))

        X = df.values
        X = np.asfortranarray(X)
        fort_predt = booster.inplace_predict(X)

        np.testing.assert_allclose(dmat_predt, arr_predt)
        np.testing.assert_allclose(df_predt, arr_predt)
        np.testing.assert_allclose(fort_predt, arr_predt)

    def test_base_margin(self):
        booster = self.booster

        base_margin = self.rng.randn(self.rows)
        from_inplace = booster.inplace_predict(data=self.X, base_margin=base_margin)

        dtrain = xgb.DMatrix(self.X, self.y, base_margin=base_margin)
        from_dmatrix = booster.predict(dtrain)
        np.testing.assert_allclose(from_dmatrix, from_inplace)

    @pytest.mark.skipif(**tm.no_pandas())
    def test_dtypes(self) -> None:
        for orig, x in np_dtypes(self.rows, self.cols):
            predt_orig = self.booster.inplace_predict(orig)
            predt = self.booster.inplace_predict(x)
            np.testing.assert_allclose(predt, predt_orig)

        # unsupported types
        for dtype in [
            np.bytes_,
            np.complex64,
            np.complex128,
        ]:
            X: np.ndarray = np.array(orig, dtype=dtype)
            with pytest.raises(ValueError):
                self.booster.inplace_predict(X)

    @pytest.mark.skipif(**tm.no_pandas())
    def test_pd_dtypes(self) -> None:
        import pandas as pd
        from pandas.api.types import is_bool_dtype

        for orig, x in pd_dtypes():
            dtypes: Union[List, pd.Series] = (
                orig.dtypes if isinstance(orig, pd.DataFrame) else [orig.dtypes]
            )
            if isinstance(orig, pd.DataFrame) and is_bool_dtype(dtypes.iloc[0]):
                continue
            y = np.arange(x.shape[0])
            Xy = xgb.DMatrix(orig, y)
            booster = xgb.train({"tree_method": "hist"}, Xy, num_boost_round=1)
            predt_orig = booster.inplace_predict(orig)
            predt = booster.inplace_predict(x)
            np.testing.assert_allclose(predt, predt_orig)



"""
Demo for using feature weight to change column sampling
=======================================================

    .. versionadded:: 1.3.0
"""

import argparse

import numpy as np
from matplotlib import pyplot as plt

import xgboost


def main(args: argparse.Namespace) -> None:
    rng = np.random.RandomState(1994)

    kRows = 4196
    kCols = 10

    X = rng.randn(kRows, kCols)
    y = rng.randn(kRows)
    fw = np.ones(shape=(kCols,))
    for i in range(kCols):
        fw[i] *= float(i)

    dtrain = xgboost.DMatrix(X, y)
    dtrain.set_info(feature_weights=fw)

    # Perform column sampling for each node split evaluation, the sampling process is
    # weighted by feature weights.
    bst = xgboost.train(
        {"tree_method": "hist", "colsample_bynode": 0.2},
        dtrain,
        num_boost_round=10,
        evals=[(dtrain, "d")],
    )
    feature_map = bst.get_fscore()

    # feature zero has 0 weight
    assert feature_map.get("f0", None) is None
    assert max(feature_map.values()) == feature_map.get("f9")

    if args.plot:
        xgboost.plot_importance(bst)
        plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--plot",
        type=int,
        default=1,
        help="Set to 0 to disable plotting the evaluation history.",
    )
    args = parser.parse_args()
    main(args)



"""
Train XGBoost with cat_in_the_dat dataset
=========================================

A simple demo for categorical data support using dataset from Kaggle categorical data
tutorial.

The excellent tutorial is at:
https://www.kaggle.com/shahules/an-overview-of-encoding-techniques

And the data can be found at:
https://www.kaggle.com/shahules/an-overview-of-encoding-techniques/data

  .. versionadded:: 1.6.0

See Also
--------
- :doc:`Tutorial </tutorials/categorical>`
- :ref:`sphx_glr_python_examples_categorical.py`
- :ref:`sphx_glr_python_examples_cat_pipeline.py`

"""

from __future__ import annotations

import os
from tempfile import TemporaryDirectory
from time import time

import pandas as pd
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split

import xgboost as xgb


def load_cat_in_the_dat() -> tuple[pd.DataFrame, pd.Series]:
    """Assuming you have already downloaded the data into `input` directory."""

    df_train = pd.read_csv("./input/cat-in-the-dat/train.csv")

    print(
        "train data set has got {} rows and {} columns".format(
            df_train.shape[0], df_train.shape[1]
        )
    )
    X = df_train.drop(["target"], axis=1)
    y = df_train["target"]

    for i in range(0, 5):
        X["bin_" + str(i)] = X["bin_" + str(i)].astype("category")

    for i in range(0, 5):
        X["nom_" + str(i)] = X["nom_" + str(i)].astype("category")

    for i in range(5, 10):
        X["nom_" + str(i)] = X["nom_" + str(i)].apply(int, base=16)

    for i in range(0, 6):
        X["ord_" + str(i)] = X["ord_" + str(i)].astype("category")

    print(
        "train data set has got {} rows and {} columns".format(X.shape[0], X.shape[1])
    )
    return X, y


params = {
    "tree_method": "hist",
    "device": "cuda",
    "n_estimators": 32,
    "colsample_bylevel": 0.7,
}


def categorical_model(X: pd.DataFrame, y: pd.Series, output_dir: str) -> None:
    """Train using builtin categorical data support from XGBoost"""
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, random_state=1994, test_size=0.2
    )
    # Be aware that the encoding for X_train and X_test are the same here. In practice,
    # we should try to use an encoder like (sklearn OrdinalEncoder) to obtain the
    # categorical values.

    # Specify `enable_categorical` to True.
    clf = xgb.XGBClassifier(
        **params,
        eval_metric="auc",
        enable_categorical=True,
        max_cat_to_onehot=1,  # We use optimal partitioning exclusively
    )
    clf.fit(X_train, y_train, eval_set=[(X_test, y_test), (X_train, y_train)])
    clf.save_model(os.path.join(output_dir, "categorical.json"))

    y_score = clf.predict_proba(X_test)[:, 1]  # proba of positive samples
    auc = roc_auc_score(y_test, y_score)
    print("AUC of using builtin categorical data support:", auc)


def onehot_encoding_model(X: pd.DataFrame, y: pd.Series, output_dir: str) -> None:
    """Train using one-hot encoded data."""
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, random_state=42, test_size=0.2
    )
    # Specify `enable_categorical` to False as we are using encoded data.
    clf = xgb.XGBClassifier(**params, eval_metric="auc", enable_categorical=False)
    clf.fit(
        X_train,
        y_train,
        eval_set=[(X_test, y_test), (X_train, y_train)],
    )
    clf.save_model(os.path.join(output_dir, "one-hot.json"))

    y_score = clf.predict_proba(X_test)[:, 1]  # proba of positive samples
    auc = roc_auc_score(y_test, y_score)
    print("AUC of using onehot encoding:", auc)


if __name__ == "__main__":
    X, y = load_cat_in_the_dat()

    with TemporaryDirectory() as tmpdir:
        start = time()
        categorical_model(X, y, tmpdir)
        end = time()
        print("Duration:categorical", end - start)

        X = pd.get_dummies(X)
        start = time()
        onehot_encoding_model(X, y, tmpdir)
        end = time()
        print("Duration:onehot", end - start)



#!/usr/bin/env python
from __future__ import annotations

import argparse
import json
import os
import re
import shutil
import subprocess
import sys
from multiprocessing import Pool, cpu_count
from time import time


def call(args: list[str]) -> tuple[int, int, str, list[str]]:
    """Subprocess run wrapper."""
    completed = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    error_msg = completed.stdout.decode("utf-8")
    # `workspace` is a name used in the CI container.  Normally we should keep the dir
    # as `xgboost`.
    matched = re.search(
        "(workspace|xgboost)/.*(ops|src|tests|include)/.*warning:",
        error_msg,
        re.MULTILINE,
    )

    if matched is None:
        return_code = 0
    else:
        return_code = 1
    return (completed.returncode, return_code, error_msg, args)


class ClangTidy:
    """clang tidy wrapper.
    Args:
      args:  Command line arguments.
          cpp_lint: Run linter on C++ source code.
          cuda_lint: Run linter on CUDA source code.
          use_dmlc_gtest: Whether to use gtest bundled in dmlc-core.
    """

    def __init__(self, args: argparse.Namespace) -> None:
        self.cpp_lint = args.cpp
        self.cuda_lint = args.cuda
        self.use_dmlc_gtest: bool = args.use_dmlc_gtest
        self.cuda_archs = args.cuda_archs.copy() if args.cuda_archs else []

        if args.tidy_version:
            self.exe = "clang-tidy-" + str(args.tidy_version)
        else:
            self.exe = "clang-tidy"

        print("Run linter on CUDA: ", self.cuda_lint)
        print("Run linter on C++:", self.cpp_lint)
        print("Use dmlc gtest:", self.use_dmlc_gtest)
        print("CUDA archs:", " ".join(self.cuda_archs))

        if not self.cpp_lint and not self.cuda_lint:
            raise ValueError("Both --cpp and --cuda are set to 0.")
        self.root_path = os.path.abspath(os.path.curdir)
        print("Project root:", self.root_path)
        self.cdb_path = os.path.join(self.root_path, "cdb")

    def __enter__(self) -> "ClangTidy":
        self.start = time()
        if os.path.exists(self.cdb_path):
            shutil.rmtree(self.cdb_path)
        self._generate_cdb()
        return self

    def __exit__(self, *args: list) -> None:
        if os.path.exists(self.cdb_path):
            shutil.rmtree(self.cdb_path)
        self.end = time()
        print("Finish running clang-tidy:", self.end - self.start)

    def _generate_cdb(self) -> None:
        """Run CMake to generate compilation database."""
        os.mkdir(self.cdb_path)
        os.chdir(self.cdb_path)
        cmake_args = [
            "cmake",
            self.root_path,
            "-GNinja",  # prevents cmake from using --option-files for include path.
            "-DCMAKE_EXPORT_COMPILE_COMMANDS=ON",
            "-DGOOGLE_TEST=ON",
            "-DCMAKE_CXX_FLAGS='-Wno-clang-diagnostic-deprecated-declarations'",
        ]
        if self.use_dmlc_gtest:
            cmake_args.append("-DUSE_DMLC_GTEST=ON")
        else:
            cmake_args.append("-DUSE_DMLC_GTEST=OFF")

        if self.cuda_lint:
            cmake_args.extend(["-DUSE_CUDA=ON", "-DUSE_NCCL=ON"])
            if self.cuda_archs:
                arch_list = ";".join(self.cuda_archs)
                cmake_args.append(f"-DCMAKE_CUDA_ARCHITECTURES={arch_list}")
        subprocess.run(cmake_args)
        os.chdir(self.root_path)

    def convert_nvcc_command_to_clang(self, command: str) -> str:
        """Convert nvcc flags to corresponding clang flags."""
        components = command.split()
        compiler: str = components[0]
        if compiler.find("nvcc") != -1:
            compiler = "clang++"
            components[0] = compiler
        # check each component in a command
        converted_components = [compiler]

        for i in range(1, len(components)):
            if components[i] == "-lineinfo":
                continue
            elif components[i] == "-fuse-ld=gold":
                continue
            elif components[i] == "-fuse-ld=lld":
                continue
            elif components[i].find("--default-stream") != -1:
                continue
            elif components[i] == "-rdynamic":
                continue
            elif components[i] == "-Xfatbin=-compress-all":
                continue
            elif components[i] == "-forward-unknown-to-host-compiler":
                continue
            elif components[i] == "-x" and components[i + 1] == "cu":
                # -x cu -> -x cuda
                converted_components.append("-x")
                converted_components.append("cuda")
                components[i + 1] = ""
                continue
            elif components[i].find("-Xcompiler") != -1:
                continue
            elif components[i].find("--expt-") != -1:
                continue
            elif components[i].find("-ccbin") != -1:
                continue
            elif components[i].find("--generate-code") != -1:
                keyword = "code=sm"
                pos = components[i].find(keyword)
                capability = components[i][
                    pos + len(keyword) + 1 : pos + len(keyword) + 3
                ]
                if pos != -1:
                    converted_components.append("--cuda-gpu-arch=sm_" + capability)
            elif components[i].find("--std=c++14") != -1:
                converted_components.append("-std=c++14")
            elif components[i].startswith("-isystem="):
                converted_components.extend(components[i].split("="))
            else:
                converted_components.append(components[i])

        converted_components.append("-isystem /usr/local/cuda/include/")

        command = ""
        for c in converted_components:
            command = command + " " + c
        command = command.strip()
        return command

    def _configure_flags(self, path: str, command: str) -> list[list[str]]:
        src = os.path.join(self.root_path, "src").replace("/", "\\/")
        include = os.path.join(self.root_path, "include").replace("/", "\\/")

        header_filter = "(" + src + "|" + include + ")"
        common_args = [
            self.exe,
            path,
            "--header-filter=" + header_filter,
            "--config-file=" + self.tidy_file,
        ]
        common_args.append("--")
        command = self.convert_nvcc_command_to_clang(command)

        command_split = command.split()[1:]  # remove clang/c++/g++
        if "-c" in command_split:
            index = command_split.index("-c")
            del command_split[index + 1]
            command_split.remove("-c")
        if "-o" in command_split:
            index = command_split.index("-o")
            del command_split[index + 1]
            command_split.remove("-o")

        common_args.extend(command_split)

        # Two passes, one for device code another for host code.
        if path.endswith("cu"):
            args = [common_args.copy(), common_args.copy()]
            args[0].append("--cuda-host-only")
            args[1].append("--cuda-device-only")
        else:
            args = [common_args.copy()]
        for a in args:
            a.append("-Wno-unused-command-line-argument")
        return args

    def _configure(self) -> list[list[str]]:
        """Load and configure compile_commands and clang_tidy."""

        def should_lint(path: str) -> bool:
            if not self.cpp_lint and path.endswith(".cc"):
                return False
            isxgb = path.find("dmlc-core") == -1
            isxgb = isxgb and (not path.startswith(self.cdb_path))
            if isxgb:
                print(path)
                return True
            return False

        cdb_file = os.path.join(self.cdb_path, "compile_commands.json")
        with open(cdb_file, "r") as fd:
            self.compile_commands = json.load(fd)

        self.tidy_file = os.path.join(self.root_path, ".clang-tidy")
        all_files = []
        for entry in self.compile_commands:
            path = entry["file"]
            if should_lint(path):
                args = self._configure_flags(path, entry["command"])
                all_files.extend(args)
        return all_files

    def run(self) -> bool:
        """Run clang-tidy."""
        all_files = self._configure()
        passed = True
        BAR = "-" * 32
        with Pool(cpu_count()) as pool:
            results = pool.map(call, all_files)
            for i, (process_status, tidy_status, msg, args) in enumerate(results):
                # Don't enforce clang-tidy to pass for now due to namespace
                # for cub in thrust is not correct.
                if tidy_status == 1:
                    passed = False
                    print(
                        BAR,
                        "\n" "Command args:",
                        " ".join(args),
                        ", ",
                        "Process return code:",
                        process_status,
                        ", ",
                        "Tidy result code:",
                        tidy_status,
                        ", ",
                        "Message:\n",
                        msg,
                        BAR,
                        "\n",
                    )
        if not passed:
            print(
                "Errors in `thrust` namespace can be safely ignored.",
                "Please address rest of the clang-tidy warnings.",
            )
        return passed


def test_tidy(args: argparse.Namespace) -> None:
    """See if clang-tidy and our regex is working correctly.  There are many subtleties
    we need to be careful. Tests here are not thorough, at least we want to guarantee
    tidy is not missing anything on the CI.

    """
    root_path = os.path.abspath(os.path.curdir)
    tidy_file = os.path.join(root_path, ".clang-tidy")
    test_file_path = os.path.join(root_path, "ops", "script", "test_tidy.cc")

    tidy_config = "--config-file=" + tidy_file
    if not args.tidy_version:
        tidy = "clang-tidy"
    else:
        tidy = "clang-tidy-" + str(args.tidy_version)
    cmd = [tidy, tidy_config, test_file_path]
    proc_code, tidy_status, error_msg, _ = call(cmd)
    if proc_code != 0 or tidy_status != 1:
        raise RuntimeError(error_msg)
    print("clang-tidy is working.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run clang-tidy.")
    parser.add_argument("--cpp", type=int, default=1)
    parser.add_argument(
        "--tidy-version",
        type=int,
        default=None,
        help="Specify the version of preferred clang-tidy.",
    )
    parser.add_argument("--cuda", type=int, default=1)
    parser.add_argument(
        "--use-dmlc-gtest",
        action="store_true",
        help="Whether to use gtest bundled in dmlc-core.",
    )
    parser.add_argument(
        "--cuda-archs", action="append", help="List of CUDA archs to build"
    )
    args = parser.parse_args()

    test_tidy(args)

    with ClangTidy(args) as linter:
        passed = linter.run()
    if not passed:
        sys.exit(1)



"""Tests for estimating the intercept."""

import json
from typing import Dict, List, Optional

import numpy as np
from scipy.special import softmax
from sklearn.datasets import (
    make_classification,
    make_multilabel_classification,
    make_regression,
)

from ..core import Booster, DMatrix, QuantileDMatrix
from ..sklearn import XGBClassifier, XGBRegressor
from ..training import train
from .updater import get_basescore
from .utils import Device, non_increasing


# pylint: disable=too-many-statements
def run_init_estimation(tree_method: str, device: Device) -> None:
    """Test for init estimation."""

    def run_reg(X: np.ndarray, y: np.ndarray) -> None:  # pylint: disable=invalid-name
        reg = XGBRegressor(
            tree_method=tree_method, max_depth=1, n_estimators=1, device=device
        )
        reg.fit(X, y, eval_set=[(X, y)])
        base_score_0 = get_basescore(reg)
        score_0 = reg.evals_result()["validation_0"]["rmse"][0]

        n_targets = 1 if y.ndim == 1 else y.shape[1]
        intercept = np.full(shape=(n_targets,), fill_value=0.5, dtype=np.float32)
        reg = XGBRegressor(
            tree_method=tree_method,
            device=device,
            max_depth=1,
            n_estimators=1,
            base_score=intercept,
        )
        reg.fit(X, y, eval_set=[(X, y)])
        base_score_1 = get_basescore(reg)
        score_1 = reg.evals_result()["validation_0"]["rmse"][0]
        assert not np.isclose(base_score_0, base_score_1).any()
        assert score_0 < score_1  # should be better

    # pylint: disable=unbalanced-tuple-unpacking
    X, y = make_regression(n_samples=4096, random_state=17)
    run_reg(X, y)
    # pylint: disable=unbalanced-tuple-unpacking
    X, y = make_regression(n_samples=4096, n_targets=3, random_state=17)
    run_reg(X, y)

    # pylint: disable=invalid-name
    def run_clf(
        X: np.ndarray, y: np.ndarray, w: Optional[np.ndarray] = None
    ) -> List[float]:
        clf = XGBClassifier(
            tree_method=tree_method, max_depth=1, n_estimators=1, device=device
        )
        if w is not None:
            clf.fit(
                X, y, sample_weight=w, eval_set=[(X, y)], sample_weight_eval_set=[w]
            )
        else:
            clf.fit(X, y, eval_set=[(X, y)])
        base_score_0 = get_basescore(clf)
        if clf.n_classes_ == 2:
            score_0 = clf.evals_result()["validation_0"]["logloss"][0]
        else:
            score_0 = clf.evals_result()["validation_0"]["mlogloss"][0]

        n_targets = 1 if y.ndim == 1 else y.shape[1]
        intercept = np.full(shape=(n_targets,), fill_value=0.5, dtype=np.float32)
        clf = XGBClassifier(
            tree_method=tree_method,
            max_depth=1,
            n_estimators=1,
            device=device,
            base_score=intercept,
        )
        if w is not None:
            clf.fit(
                X, y, sample_weight=w, eval_set=[(X, y)], sample_weight_eval_set=[w]
            )
        else:
            clf.fit(X, y, eval_set=[(X, y)])
        base_score_1 = get_basescore(clf)
        if clf.n_classes_ == 2:
            score_1 = clf.evals_result()["validation_0"]["logloss"][0]
        else:
            score_1 = clf.evals_result()["validation_0"]["mlogloss"][0]
        assert not np.isclose(base_score_0, base_score_1).any()
        assert score_0 < score_1 + 1e-4  # should be better

        return base_score_0

    # pylint: disable=unbalanced-tuple-unpacking
    X, y = make_classification(n_samples=4096, random_state=17)
    run_clf(X, y)
    X, y = make_multilabel_classification(
        n_samples=4096, n_labels=3, n_classes=5, random_state=17
    )
    run_clf(X, y)

    # Extra tests for the classifier.
    X, y = make_classification(
        n_samples=4096, random_state=17, n_classes=5, n_informative=20, n_redundant=0
    )
    intercept = run_clf(X, y)
    # un-transformed intercept sums to 0, as a convention.
    np.testing.assert_allclose(np.sum(softmax(intercept)), 1.0)
    np.testing.assert_allclose(np.sum(intercept), 0.0, atol=1e-6)

    assert np.all(softmax(intercept) > 0)
    np_int = (
        np.histogram(
            y, bins=np.concatenate([np.unique(y), np.array([np.finfo(np.float32).max])])
        )[0]
        / y.shape[0]
    )
    np.testing.assert_allclose(softmax(intercept), np_int, rtol=1e-6)

    rng = np.random.default_rng(1994)
    w = rng.uniform(low=0, high=1, size=(y.shape[0],))
    intercept = run_clf(X, y, w)
    np.testing.assert_allclose(np.sum(softmax(intercept)), 1.0)
    assert np.all(softmax(intercept) > 0)


# pylint: disable=too-many-locals
def run_adaptive(tree_method: str, weighted: bool, device: Device) -> None:
    """Test for adaptive trees."""
    rng = np.random.RandomState(1994)
    from sklearn.utils import stats

    n_samples = 256
    X, y = make_regression(  # pylint: disable=unbalanced-tuple-unpacking
        n_samples, 16, random_state=rng
    )
    if weighted:
        w = rng.normal(size=n_samples)
        w -= w.min()
        Xy = DMatrix(X, y, weight=w)

        kwargs = {"percentile_rank": 50}
        base_score = stats._weighted_percentile(  # pylint: disable=protected-access
            y, w, **kwargs
        )
    else:
        Xy = DMatrix(X, y)
        base_score = np.median(y)

    # Check the base score is expected.
    booster_0 = train(
        {
            "tree_method": tree_method,
            "base_score": base_score,
            "objective": "reg:absoluteerror",
            "device": device,
        },
        Xy,
        num_boost_round=1,
    )
    booster_1 = train(
        {
            "tree_method": tree_method,
            "objective": "reg:absoluteerror",
            "device": device,
        },
        Xy,
        num_boost_round=1,
    )
    config_0 = json.loads(booster_0.save_config())
    config_1 = json.loads(booster_1.save_config())

    assert get_basescore(config_0) == get_basescore(config_1)

    # check the base score is correctly serialized.
    raw_booster = booster_1.save_raw(raw_format="ubj")
    booster_2 = Booster(model_file=raw_booster)
    config_2 = json.loads(booster_2.save_config())
    assert get_basescore(config_1) == get_basescore(config_2)

    # check we can override the base score.
    booster_0 = train(
        {
            "tree_method": tree_method,
            "base_score": base_score + 1.0,
            "objective": "reg:absoluteerror",
            "device": device,
        },
        Xy,
        num_boost_round=1,
    )
    config_0 = json.loads(booster_0.save_config())
    np.testing.assert_allclose(
        get_basescore(config_0), np.asarray(get_basescore(config_1)) + 1
    )

    # check we can use subsampling.
    evals_result: Dict[str, Dict[str, list]] = {}
    train(
        {
            "tree_method": tree_method,
            "device": device,
            "objective": "reg:absoluteerror",
            "subsample": 0.8,
            "eta": 1.0,
        },
        Xy,
        num_boost_round=10,
        evals=[(Xy, "Train")],
        evals_result=evals_result,
    )
    mae = evals_result["Train"]["mae"]
    assert mae[-1] < 20.0
    assert non_increasing(mae)


def run_exp_family(device: Device) -> None:
    """Exp family has a closed solution."""
    X, y = make_classification(n_samples=128, n_classes=2, weights=[0.8, 0.2])
    Xy = QuantileDMatrix(X, y)
    clf = train(
        {"objective": "binary:logistic", "device": device}, Xy, num_boost_round=1
    )
    reg = train({"objective": "reg:logistic", "device": device}, Xy, num_boost_round=1)
    clf1 = train(
        {"objective": "binary:logitraw", "device": device}, Xy, num_boost_round=1
    )
    # The base score stored in the booster model is un-transformed
    np.testing.assert_allclose([get_basescore(m) for m in (reg, clf, clf1)], y.mean())

    X, y = make_classification(weights=[0.8, 0.2], random_state=2025)
    clf = train(
        {"objective": "binary:logistic", "scale_pos_weight": 4.0, "device": device},
        QuantileDMatrix(X, y),
        num_boost_round=1,
    )
    score = get_basescore(clf)
    np.testing.assert_allclose(score, 0.5, rtol=1e-3)


def run_logistic_degenerate(device: Device) -> None:
    """Test https://github.com/dmlc/xgboost/issues/11499 ."""

    def run(v: float) -> None:
        dtrain = DMatrix(np.asarray([[1.0], [1.0]]), label=[v, v])
        bst = train(
            {"objective": "binary:logistic", "device": device},
            dtrain,
            1,
        )
        intercept = get_basescore(bst)
        assert intercept[0] == v

    run(0.0)
    run(1.0)



"""
Upload an artifact to an S3 bucket for later use
Note. This script takes in all inputs via environment variables
      except the path to the artifact(s).
"""

import argparse
import os
import subprocess
import sys
from pathlib import Path
from urllib.parse import SplitResult, urlsplit, urlunsplit


def resolve(x: Path) -> Path:
    return x.expanduser().resolve()


def path_equals(a: Path, b: Path) -> bool:
    return resolve(a) == resolve(b)


def compute_s3_url(*, s3_bucket: str, prefix: str, artifact: str) -> str:
    if prefix == "":
        return f"s3://{s3_bucket}/{artifact}"
    return f"s3://{s3_bucket}/{prefix}/{artifact}"


def aws_s3_upload(*, src: Path, dest: str, make_public: bool) -> None:
    cli_args = ["aws", "s3", "cp", "--no-progress", str(src), dest]
    if make_public:
        cli_args.extend(["--acl", "public-read"])
    print(" ".join(cli_args))
    subprocess.run(
        cli_args,
        check=True,
        encoding="utf-8",
    )


def aws_s3_download(*, src: str, dest_dir: Path) -> None:
    cli_args = ["aws", "s3", "cp", "--no-progress", src, str(dest_dir)]
    print(" ".join(cli_args))
    subprocess.run(
        cli_args,
        check=True,
        encoding="utf-8",
    )


def aws_s3_download_with_wildcard(*, src: str, dest_dir: Path) -> None:
    parsed_src = urlsplit(src)
    src_dir = urlunsplit(
        SplitResult(
            scheme="s3",
            netloc=parsed_src.netloc,
            path=os.path.dirname(parsed_src.path),
            query="",
            fragment="",
        )
    )
    src_glob = os.path.basename(parsed_src.path)
    cli_args = [
        "aws",
        "s3",
        "cp",
        "--recursive",
        "--no-progress",
        "--exclude",
        "'*'",
        "--include",
        src_glob,
        src_dir,
        str(dest_dir),
    ]
    print(" ".join(cli_args))
    subprocess.run(
        cli_args,
        check=True,
        encoding="utf-8",
    )


def upload(*, args: argparse.Namespace) -> None:
    print(f"Uploading artifacts to prefix {args.prefix}...")
    for artifact in args.artifacts:
        artifact_path = Path(artifact)
        s3_url = compute_s3_url(
            s3_bucket=args.s3_bucket, prefix=args.prefix, artifact=artifact_path.name
        )
        aws_s3_upload(src=artifact_path, dest=s3_url, make_public=args.make_public)


def download(*, args: argparse.Namespace) -> None:
    print(f"Downloading artifacts from prefix {args.prefix}...")
    dest_dir = Path(args.dest_dir)
    print(f"mkdir -p {str(dest_dir)}")
    dest_dir.mkdir(parents=True, exist_ok=True)
    for artifact in args.artifacts:
        s3_url = compute_s3_url(
            s3_bucket=args.s3_bucket, prefix=args.prefix, artifact=artifact
        )
        if "*" in artifact:
            aws_s3_download_with_wildcard(src=s3_url, dest_dir=dest_dir)
        else:
            aws_s3_download(src=s3_url, dest_dir=dest_dir)


if __name__ == "__main__":
    # Ensure that the current working directory is the project root
    if not (Path.cwd() / "ops").is_dir() or not path_equals(
        Path(__file__).parent.parent, Path.cwd() / "ops"
    ):
        x = Path(__file__).name
        raise RuntimeError(f"Script {x} must be run at the project's root directory")

    root_parser = argparse.ArgumentParser()
    subparser_factory = root_parser.add_subparsers(required=True, dest="command")
    parsers = {}
    for command in ["upload", "download"]:
        parsers[command] = subparser_factory.add_parser(command)
        parsers[command].add_argument(
            "--s3-bucket",
            type=str,
            required=True,
            help="Name of the S3 bucket to store the artifact",
        )
        parsers[command].add_argument(
            "--prefix",
            type=str,
            required=True,
            help=(
                "Where the artifact(s) would be stored. The artifact(s) will be stored at "
                "s3://[s3-bucket]/[prefix]/[filename]."
            ),
        )
        parsers[command].add_argument(
            "artifacts",
            type=str,
            nargs="+",
            metavar="artifact",
            help=f"Artifact(s) to {command}",
        )

    parsers["upload"].add_argument(
        "--make-public", action="store_true", help="Make artifact publicly accessible"
    )
    parsers["download"].add_argument(
        "--dest-dir", type=str, required=True, help="Where to download artifact(s)"
    )

    if len(sys.argv) == 1:
        print("1. Upload artifact(s)")
        parsers["upload"].print_help()
        print("\n2. Download artifact(s)")
        parsers["download"].print_help()
        sys.exit(1)

    parsed_args = root_parser.parse_args()
    if parsed_args.command == "upload":
        upload(args=parsed_args)
    elif parsed_args.command == "download":
        download(args=parsed_args)



"""Xgboost pyspark integration submodule for helper functions."""

# pylint: disable=fixme

import inspect
import logging
import sys
from threading import Thread
from typing import Any, Callable, Dict, Optional, Set, Type, Union

import pyspark
from pyspark import BarrierTaskContext, TaskContext
from pyspark.sql import SparkSession

from ..collective import CommunicatorContext as CCtx
from ..collective import Config
from ..collective import _Args as CollArgs
from ..collective import _ArgVals as CollArgsVals
from ..core import Booster
from ..sklearn import XGBModel
from ..tracker import RabitTracker


def get_class_name(cls: Type) -> str:
    """Return the class name."""
    return f"{cls.__module__}.{cls.__name__}"


def _get_default_params_from_func(
    func: Callable, unsupported_set: Set[str]
) -> Dict[str, Any]:
    """Returns a dictionary of parameters and their default value of function fn.  Only
    the parameters with a default value will be included.

    """
    sig = inspect.signature(func)
    filtered_params_dict = {}
    for parameter in sig.parameters.values():
        # Remove parameters without a default value and those in the unsupported_set
        if (
            parameter.default is not parameter.empty
            and parameter.name not in unsupported_set
        ):
            filtered_params_dict[parameter.name] = parameter.default
    return filtered_params_dict


class CommunicatorContext(CCtx):
    """Context with PySpark specific task ID."""

    def __init__(self, context: BarrierTaskContext, **args: CollArgsVals) -> None:
        args["dmlc_task_id"] = str(context.partitionId())
        super().__init__(**args)


def _start_tracker(host: str, n_workers: int, port: int = 0) -> CollArgs:
    """Start Rabit tracker with n_workers"""
    args: CollArgs = {"n_workers": n_workers}
    tracker = RabitTracker(n_workers=n_workers, host_ip=host, sortby="task", port=port)
    tracker.start()
    thread = Thread(target=tracker.wait_for)
    thread.daemon = True
    thread.start()
    args.update(tracker.worker_args())
    return args


def _get_rabit_args(conf: Config, n_workers: int) -> CollArgs:
    """Get rabit context arguments to send to each worker."""
    assert conf.tracker_host_ip is not None
    port = 0 if conf.tracker_port is None else conf.tracker_port
    env = _start_tracker(conf.tracker_host_ip, n_workers, port)
    return env


def _get_host_ip(context: BarrierTaskContext) -> str:
    """Gets the hostIP for Spark. This essentially gets the IP of the first worker."""
    task_ip_list = [info.address.split(":")[0] for info in context.getTaskInfos()]
    return task_ip_list[0]


def get_logger(name: str, level: Optional[Union[str, int]] = None) -> logging.Logger:
    """Gets a logger by name, or creates and configures it for the first time."""
    logger = logging.getLogger(name)
    if level is not None:
        logger.setLevel(level)
    else:
        # Default to info if not set.
        if logger.level == logging.NOTSET:
            logger.setLevel(logging.INFO)
    # If the logger is configured, skip the configure
    if not logger.handlers and not logging.getLogger().handlers:
        handler = logging.StreamHandler(sys.stderr)
        formatter = logging.Formatter(
            "%(asctime)s %(levelname)s %(name)s: %(funcName)s %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger


def get_logger_level(name: str) -> Optional[int]:
    """Get the logger level for the given log name"""
    logger = logging.getLogger(name)
    return None if logger.level == logging.NOTSET else logger.level


def _get_max_num_concurrent_tasks(spark_session: SparkSession) -> int:
    """Gets the current max number of concurrent tasks."""

    # In Spark Connect, we cannot easily get the max number of concurrent tasks
    # from the client side without accessing internal APIs or executing a task.
    # For now, we return a large number to skip the check.
    if _is_connect(spark_session):
        return sys.maxsize

    # pylint: disable=protected-access
    return spark_session.sparkContext._jsc.sc().maxNumConcurrentTasks(
        spark_session.sparkContext._jsc.sc()
        .resourceProfileManager()
        .resourceProfileFromId(0)
    )


def _is_connect(spark_session: SparkSession) -> bool:
    try:
        return isinstance(spark_session, pyspark.sql.connect.session.SparkSession)
    except AttributeError:
        return False


def _is_local(spark_session: SparkSession) -> bool:
    """Whether it is Spark local mode"""
    # In Spark Connect, we check the spark.master configuration if available.
    # Note: This might not be accurate if spark.master is not set in RuntimeConfig.
    master = spark_session.conf.get("spark.master", None)
    return master is not None and (master == "local" or master.startswith("local["))


def _get_gpu_id(task_context: TaskContext) -> int:
    """Get the gpu id from the task resources"""
    if task_context is None:
        # This is a safety check.
        raise RuntimeError("_get_gpu_id should not be invoked from driver side.")
    resources = task_context.resources()
    if "gpu" not in resources:
        raise RuntimeError(
            "Couldn't get the gpu id, Please check the GPU resource configuration"
        )
    # return the first gpu id.
    return int(resources["gpu"].addresses[0].strip())


def deserialize_xgb_model(
    model: str, xgb_model_creator: Callable[[], XGBModel]
) -> XGBModel:
    """
    Deserialize an xgboost.XGBModel instance from the input model.
    """
    xgb_model = xgb_model_creator()
    xgb_model.load_model(bytearray(model.encode("utf-8")))
    return xgb_model


def serialize_booster(booster: Booster) -> str:
    """
    Serialize the input booster to a string.

    Parameters
    ----------
    booster:
        an xgboost.core.Booster instance
    """
    return booster.save_raw("json").decode("utf-8")


def deserialize_booster(model: str) -> Booster:
    """
    Deserialize an xgboost.core.Booster from the input ser_model_string.
    """
    booster = Booster()
    booster.load_model(bytearray(model.encode("utf-8")))
    return booster


def use_cuda(device: Optional[str]) -> bool:
    """Whether xgboost is using CUDA workers."""
    return device in ("cuda", "gpu")



"""
Example of training with Dask on CPU
====================================

"""

from dask import array as da
from dask.distributed import Client, LocalCluster

from xgboost import dask as dxgb
from xgboost.dask import DaskDMatrix


def main(client: Client) -> None:
    # generate some random data for demonstration
    m = 100000
    n = 100
    rng = da.random.default_rng(1)
    X = rng.normal(size=(m, n), chunks=(10000, -1))
    y = X.sum(axis=1)

    # DaskDMatrix acts like normal DMatrix, works as a proxy for local
    # DMatrix scatter around workers.
    dtrain = DaskDMatrix(client, X, y)

    # Use train method from xgboost.dask instead of xgboost.  This
    # distributed version of train returns a dictionary containing the
    # resulting booster and evaluation history obtained from
    # evaluation metrics.
    output = dxgb.train(
        client,
        {"verbosity": 1, "tree_method": "hist"},
        dtrain,
        num_boost_round=4,
        evals=[(dtrain, "train")],
    )
    bst = output["booster"]
    history = output["history"]

    # you can pass output directly into `predict` too.
    prediction = dxgb.predict(client, bst, dtrain)
    print("Evaluation history:", history)
    print("Error:", da.sqrt((prediction - y) ** 2).mean().compute())


if __name__ == "__main__":
    # or use other clusters for scaling
    with LocalCluster(n_workers=7, threads_per_worker=4) as cluster:
        with Client(cluster) as client:
            main(client)



"""Shared fixtures for Dask tests."""

from __future__ import annotations

import os
from typing import Any, Dict, Generator

import pytest
from distributed import Client, LocalCluster


@pytest.fixture(scope="session")
def client_kwargs(request: pytest.FixtureRequest) -> Dict[str, Any]:
    n_threads = os.cpu_count()
    assert n_threads is not None
    kwargs: Dict[str, Any] = {
        "n_workers": 2,
        "threads_per_worker": max(1, n_threads // 2),
        "dashboard_address": ":0",
    }
    if hasattr(request, "param"):
        kwargs.update(request.param)
    return kwargs


@pytest.fixture(scope="session")
def client(client_kwargs: Dict[str, Any]) -> Generator[Client, None, None]:
    with LocalCluster(**client_kwargs) as dask_cluster:
        with Client(dask_cluster) as dask_client:
            yield dask_client


@pytest.fixture(autouse=True)
def client_as_current(request: pytest.FixtureRequest) -> Generator[None, None, None]:
    for name in ("client", "client_one_worker"):
        if name in request.fixturenames:
            dask_client = request.getfixturevalue(name)
            with dask_client.as_current():
                yield
            return
    yield


@pytest.fixture(scope="session")
def client_one_worker() -> Generator[Client, None, None]:
    n_threads = os.cpu_count()
    assert n_threads is not None
    with LocalCluster(
        n_workers=1, threads_per_worker=max(1, n_threads), dashboard_address=":0"
    ) as dask_cluster:
        with Client(dask_cluster) as dask_client:
            yield dask_client
